"""Console script for pixelator (adapterqc).

Copyright © 2022 Pixelgen Technologies AB.
"""

import click

from pixelator.common.utils import (
    create_output_stage_dir,
    get_extension,
    get_sample_name,
    log_step_start,
    sanity_check_inputs,
    timer,
    write_parameters_file,
)
from pixelator.mpx.cli.common import design_option, logger, output_option
from pixelator.mpx.qc import adapter_qc_fastq


@click.command(
    "adapterqc",
    short_help=(
        "process pixel data (FASTQ) to check for the presence of PBS1/2 sequences"
    ),
    options_metavar="<options>",
)
@click.argument(
    "fastq_file",
    nargs=1,
    required=True,
    type=click.Path(exists=True),
    metavar="FASTQ",
)
@click.option(
    "--mismatches",
    default=0.1,
    required=False,
    type=click.FloatRange(0.0, 0.9),
    show_default=True,
    help="The number of mismatches allowed (in percentage)",
)
@output_option
@design_option
@click.pass_context
@timer
def adapterqc(
    ctx,
    fastq_file,
    mismatches,
    output,
    design,
):
    """Check for the presence of PBS1/2 sequences in FASTQ input files."""
    from pixelator.mpx.config.config_instance import config

    input_files = [fastq_file]
    # log input parameters
    log_step_start(
        "adapterqc",
        input_files=input_files,
        output=output,
        mismatches=mismatches,
        design=design,
    )

    # some basic sanity check on the input files
    sanity_check_inputs(input_files, allowed_extensions=("fastq.gz", "fq.gz"))

    # update arguments from config file (CLI arguments have priority)
    try:
        amplicon = config.get_assay(design).get_region_by_id("amplicon")
        pbs1 = amplicon.get_region_by_id("pbs-1").get_sequence()
        pbs2 = amplicon.get_region_by_id("pbs-2").get_sequence()
    except KeyError as exc:
        raise click.ClickException(f"Parsing attribute from config file {exc}")

    # create output folder if it does not exist
    adapterqc_output = create_output_stage_dir(output, "adapterqc")

    # run cutadapt (adapter mode) using parallel processing
    msg = f"Processing {fastq_file} with cutadapt (adapter mode)"
    logger.info(msg)

    clean_name = get_sample_name(fastq_file)
    extension = get_extension(fastq_file)
    output_file = adapterqc_output / f"{clean_name}.processed.{extension}"
    failed_file = adapterqc_output / f"{clean_name}.failed.{extension}"
    json_file = adapterqc_output / f"{clean_name}.report.json"

    write_parameters_file(
        ctx,
        adapterqc_output / f"{clean_name}.meta.json",
        command_path="pixelator single-cell-mpx adapterqc",
    )

    adapter_qc_fastq(
        input=fastq_file,
        output=str(output_file),
        failed=str(failed_file),
        report=str(json_file),
        mismatches=mismatches,
        pbs1=pbs1,
        pbs2=pbs2,
        cores=0,  # cutadapt will choose the optimal number
        verbose=ctx.obj["VERBOSE"],
    )



"""Functions for statistics/math.

This module contains functions related to various useful statistics/math

Copyright © 2023 Pixelgen Technologies AB.
"""

import logging
from typing import List, Literal, Union

import numpy as np
import pandas as pd
from scipy.stats import mannwhitneyu
from sklearn.decomposition import PCA
from sklearn.mixture import GaussianMixture
from sklearn.preprocessing import StandardScaler

logger = logging.getLogger(__name__)


def clr_transformation(
    df: pd.DataFrame,
    axis: Literal[0, 1] = 0,
    non_negative: bool = True,
) -> pd.DataFrame:
    """Transform antibody counts data with CLR (centered log ratio).

    This function performs a CLR (centered log ratio) transformation on the
    provided dataframe containing antibody counts. The CLR transformation
    divides the counts by the geometric mean and then applies a log
    transformation. Alternatively, it can log-transform the counts first and
    then subtract the geometric mean (log), centering the transformed counts
    around zero (which may include negative values).

    Args:
        df (pd.DataFrame): The dataframe of antibody counts.
        axis (Literal[0, 1], optional): The axis on which to apply the
            transformation. `axis=0` applies the transformation by columns
            (antibody), and `axis=1` applies it by rows (component). Defaults
            to 0.
        non_negative (bool, optional): If `True`, the non-negative CLR
            transformation is used. If `False`, the zero-centered CLR
            transformation is used. Defaults to True.

    Raises:
        AssertionError: If the input axis is not 0 or 1.

    Returns:
        pd.DataFrame: A dataframe with the antibody counts transformed.

    References:
        https://en.wikipedia.org/wiki/Compositional_data#Center_logratio_transform

    """
    if axis not in [0, 1]:
        raise AssertionError("Axis is required to be 0 or 1")

    logger.debug(
        "Computing CLR transformation for antibody counts with %i nodes and %i markers",
        df.shape[0],
        df.shape[1],
    )

    geometric_mean_log = np.mean(np.log1p(df), axis=axis)
    dim = 1 if axis == 0 else 0
    if non_negative:
        # using the definition of clr from Wikipedia:
        # clr(X) = [log(x1 / g(X)), ..., log(xn / g(X))]
        # where X is a vector, and g(X) is the geometric mean
        # this is equivalent to clr(X) = log(X / g(X))
        # this definition is used below with a pseudocount of 1:
        clr_df = np.log1p(df.div(np.exp(geometric_mean_log), axis=dim))
    else:
        # since log(A/B) = log(A) - log(B) we can rewrite the term above to:
        # clr(X) = [log(x1) - log(g(X)), ..., log(xn) - log(g(X))]
        # this definition is used below with a pseudocount of 1:
        clr_df = np.log1p(df).subtract(geometric_mean_log, axis=dim)

    logger.debug("CLR transformation computed")
    return clr_df


def correct_pvalues(pvalues: np.ndarray) -> np.ndarray:
    """Correct a series of p-values using the Benjamini-Hochberg method.

    An outline of the method can be found here:
    https://en.wikipedia.org/wiki/False_discovery_rate#Benjamini%E2%80%93Hochberg_procedure

    Args:
        pvalues (np.ndarray): An array of p-values to adjust.

    Returns:
        np.ndarray: The array of adjusted p-values in the same order as the input array.

    """
    # Most descriptions of the BH method states that p-values should
    # first be ordered in ascending order an ranked, however doing so
    # requires reversing the sort-order multiple times. For this reason
    # we sort them in descending order instead.
    descending_order_idx = pvalues.argsort()[::-1]
    original_order_idx = descending_order_idx.argsort()

    p_values_in_desc_order = pvalues[descending_order_idx]
    total_nbr_of_p_values = float(len(pvalues))
    descending_rank = np.arange(len(pvalues), 0, -1)

    q = np.where(
        np.isnan(p_values_in_desc_order),
        np.nan,
        np.fmin.accumulate(
            (p_values_in_desc_order * total_nbr_of_p_values / descending_rank)
        ),
    )
    q[q > 1] = 1
    return q[original_order_idx]


def log1p_transformation(df: pd.DataFrame) -> pd.DataFrame:
    """Transform antibody counts using the log1p function.

    This function applies the natural logarithm of (1 + x) to the count of each
    marker or component, element-wise.

    Args:
        df (pd.DataFrame): The dataframe of antibody counts (antibodies as columns).

    Returns:
        pd.DataFrame: A dataframe with the counts normalized.

    """
    logger.debug(
        (
            "Computing LOG1P normalization for antibody counts"
            " with %i nodes and %i markers"
        ),
        df.shape[0],
        df.shape[1],
    )

    log1p_df = df.transform(np.log1p)

    logger.debug("LOG1P transformation computed")
    return log1p_df


def rate_diff_transformation(df: pd.DataFrame) -> pd.DataFrame:
    """Transform antibody counts as deviation from an expected baseline distribution.

    The baseline distribution refers to a fixed ratio of different antibody types
    in each node. For example, if 10% of antibodies are HLA-ABC, in a node with
    120 antibodies, the expected count is 12. If the actual count is 8, the
    transformation for HLA-ABC in this node will be -4.

    Args:
        df (pd.DataFrame): The dataframe of raw antibody counts (antibodies as columns).

    Returns:
        pd.DataFrame: A dataframe with the counts difference from expected values.

    """
    antibody_counts_per_node = df.sum(axis=1)
    antibody_rates = df.sum(axis=0)
    antibody_rates = antibody_rates / antibody_rates.sum()

    expected_counts = antibody_counts_per_node.to_frame() @ antibody_rates.to_frame().T
    return df - expected_counts


def rel_normalization(df: pd.DataFrame, axis: Literal[0, 1] = 0) -> pd.DataFrame:
    """Normalize antibody counts to the relative amount per marker or component.

    This function normalizes antibody counts using relative counts, where the
    count of each marker or component is divided by its total sum. Use `axis=0`
    to apply the normalization by column (antibody) and `axis=1` to apply it by
    row (component).

    Args:
        df (pd.DataFrame): The dataframe of antibody counts (antibodies as columns).
        axis (Literal[0, 1]): The axis on which to apply the normalization.
            `axis=0` applies normalization by columns, and `axis=1` applies it by rows.

    Raises:
        AssertionError: If the input axis is not 0 or 1.

    Returns:
        pd.DataFrame: A dataframe with the counts normalized.

    """
    if axis not in [0, 1]:
        raise AssertionError("Axis is required to be 0 or 1")

    logger.debug(
        "Computing REL normalization for antibody counts with %i nodes and %i markers",
        df.shape[0],
        df.shape[1],
    )

    norm_df = df.div(df.sum(axis=axis), axis=1 if axis == 0 else 0)

    logger.debug("REL normalization computed")
    return norm_df


def wilcoxon_test(
    df: pd.DataFrame,
    reference: str,
    target: str,
    contrast_column: str,
    value_column: str,
) -> pd.Series:
    """Perform a Wilcoxon rank-sum test between two groups.

    Args:
        df (pd.DataFrame): The dataframe containing the data.
        reference (str): Name of the reference group in the contrast column.
        target (str): Name of the target group in the contrast column.
        contrast_column (str): Name of the column containing the group information.
        value_column (str): Name of the column containing the values to compare.

    Returns:
        pd.Series: A series containing the test statistic, p-value, and median difference.

    """
    reference_df = df.loc[df[contrast_column] == reference, :]
    target_df = df.loc[df[contrast_column] == target, :]

    if reference_df.empty or target_df.empty:
        return pd.Series({"stat": 0, "p_value": 1, "median_difference": 0})

    estimate = np.median(
        target_df[value_column].to_numpy()[:, None]
        - reference_df[value_column].to_numpy()
    )

    stat, p_value = mannwhitneyu(
        x=reference_df[value_column],
        y=target_df[value_column],
        alternative="two-sided",
    )

    return pd.Series({"stat": stat, "p_value": p_value, "median_difference": estimate})


def _regress_out_confounder(pheno, exprs, rcond=1e-8):
    """Linear regression to remove confounding factors from abundance data."""
    design_matrix = np.column_stack((np.ones((len(pheno), 1)), pheno))
    coefficients, res, rank, s = np.linalg.lstsq(design_matrix, exprs, rcond=rcond)
    beta = coefficients[1:]  # remove intercept term
    return exprs - design_matrix[:, 1:].dot(beta)


def _get_background_abundance(dataframe: pd.DataFrame, axis=0):
    """Estimate the background abundance of a marker or component.

    Fit a double gaussian distribution to the abundance data and return the
    mean of the first gaussian as an estimation of the background level.
    """
    background = pd.Series(index=dataframe.index if axis == 0 else dataframe.columns)
    scores = pd.Series(index=dataframe.index if axis == 0 else dataframe.columns)
    gmm = GaussianMixture(n_components=2, max_iter=1000, random_state=0)
    if axis not in {0, 1}:
        raise ValueError(f"Axis was {axis}. Must be 0 or 1")
    ax_iter = dataframe.index if axis == 0 else dataframe.columns
    for i in ax_iter:
        current_axis = dataframe.loc[i, :] if axis == 0 else dataframe.loc[:, i]
        gmm = gmm.fit(current_axis.to_frame())
        background[i] = np.min(gmm.means_)
        scores[i] = np.abs(gmm.means_[1] - gmm.means_[0]) / np.sum(gmm.covariances_)
    return background, scores


def dsb_normalize(
    raw_abundance: pd.DataFrame, isotype_controls: Union[List, None] = None
):
    """Normalize abundance data using the empty-droplet-free method.

    This method is implemented as described in Mulè et al.'s dsb package.
    The normalization steps are:
    1. Log1p transformation.
    2. Remove background abundance per marker.
    3. Regularize abundance per component.

    Args:
        raw_abundance (pd.DataFrame): The raw abundance count data.
        isotype_controls (Union[List, None]): List of isotype controls.

    Raises:
        ValueError: If no isotype controls are provided.

    Returns:
        pd.DataFrame: Normalized abundance data.

    References:
        Integrating population and single-cell variations in vaccine responses
        identifies a naturally adjuvanted human immune setpoint,
        Matthew P. Mulè et al., Immunity, 2024,
        https://doi.org/10.1016/j.immuni.2024.04.009

    """
    log_abundance = np.log1p(raw_abundance)
    marker_background, _ = _get_background_abundance(log_abundance, axis=1)
    log_abundance = log_abundance - marker_background
    component_background, _ = _get_background_abundance(log_abundance, axis=0)

    if isotype_controls is not None:
        control_signals = log_abundance.loc[:, isotype_controls]
        control_signals["component_background"] = component_background
        control_signals = StandardScaler().fit_transform(control_signals)
        pheno = PCA(n_components=1).fit_transform(control_signals)
    else:
        raise ValueError("At least one isotype control must be provided.")

    normalized_abundance = _regress_out_confounder(pheno, log_abundance)

    return normalized_abundance



"""Common functions and utilities for Pixelator.

Copyright © 2022 Pixelgen Technologies AB.
"""

from __future__ import annotations

import collections.abc
import gzip
import itertools
import json
import logging
import multiprocessing
import re
import tempfile
import textwrap
import time
import typing
from concurrent.futures import ProcessPoolExecutor
from functools import wraps
from logging.handlers import SocketHandler
from multiprocessing.pool import Pool
from pathlib import Path, PurePath
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generator,
    Iterable,
    List,
    Literal,
    Optional,
    Sequence,
    Set,
    Union,
)

import click
import numpy as np
import pandas as pd

from pixelator.common.types import PathType

# Avoid a circular dependency
if TYPE_CHECKING:
    from pixelator.common.config import AntibodyPanel

logger = logging.getLogger(__name__)

# this tr table is used to complement DNA sequences
_TRTABLE = str.maketrans("GTACN", "CATGN")


def build_barcodes_file(
    panel: AntibodyPanel, anchored: bool, rev_complement: bool
) -> str:
    """Create a FASTA file of barcodes from a panel dataframe.

    The FASTA file will have the marker id as
    name and the barcode sequence as sequence. The parameter
    rev_complement control if sequence needs to be in reverse
    complement form or not. When anchored is true a dollar sign
    ($) will be added at the end of the sequences to make them
    anchored in cutadapt.

    :param panel: an Antibody panel object
    :param anchored: make the sequences anchored if True
    :param rev_complement: reverse complement the sequence column
                           if True
    :returns str: a path to the barcodes file
    """
    logger.debug("Creating barcodes file from antibody panel")

    fd = tempfile.NamedTemporaryFile(suffix=".fa", delete=False)
    logger.info("Barcodes file saved in %s", fd.name)
    with open(fd.name, "w") as fh:
        for _, row in panel.df.iterrows():
            marker_id = row["marker_id"]
            sequence = row["sequence"]
            fh.write(f">{sequence} [marker_id={marker_id}]\n")
            seq = reverse_complement(sequence) if rev_complement else sequence
            if anchored:
                seq += "$"
            fh.write(f"{seq}\n")

    logger.debug("Barcodes file from antibody panel created")
    return fd.name


def click_echo(msg: str, multiline: bool = False):
    """Print a line to the console with optional long-line wrapping.

    :param msg: the message to print
    :param multiline: True to use text wrapping or False otherwise (default)
    """
    if multiline:
        click.echo(textwrap.fill(textwrap.dedent(msg), width=100))
    else:
        click.echo(msg)


def create_output_stage_dir(root: PathType, name: str) -> Path:
    """Create a new subfolder with `name` under the given `root` directory.

    :param root: the parent directory
    :param name: the name of the directory to create
    :returns Path: the created folder (Path)
    """
    output = Path(root) / name
    if not output.is_dir():
        output.mkdir(parents=True)
    return output


def batched(iterable, n):
    """Batch data into tuples of length n. The last batch may be shorter.

    Taken from python itertools recipes.
    """
    if n < 1:
        raise ValueError("n must be at least one")
    it = iter(iterable)
    while batch := tuple(itertools.islice(it, n)):
        yield batch


def get_extension(filename: PathType, len_ext: int = 2) -> str:
    """Extract file extensions from a filename.

    :param filename: the file name
    :param len_ext: the number of expected extensions parts
        e.g.: fq.gz gives len_ext=2
    :returns str: the file extension (str)
    """
    return "".join(PurePath(filename).suffixes[-len_ext:]).lstrip(".")


def get_sample_name(filename: PathType) -> str:
    """Extract the sample name from a sample's filename.

    The sample name is expected to be from the start of the filename until
    the first dot.

    :param filename: path to the file
    :returns str: the sample name
    """
    return Path(filename).stem.split(".")[0]


def group_input_reads(
    inputs: Sequence[PathType], input1_pattern: str, input2_pattern: str
) -> Dict[str, List[Path]]:
    """Group input files by read pairs and sample id.

    :param inputs: list of input files
    :param input1_pattern: pattern to match read1 files
    :param input2_pattern: pattern to match read2 files
    :raises ValueError: if the number of reads for a sample is more than 2
    :returns Dict[str, List[Path]]: a dictionary with the grouped reads
    """

    def group_fn(s):
        sn = get_sample_name(s)
        return sn.replace(input1_pattern, "").replace(input2_pattern, "")

    inputs = sorted(inputs, key=group_fn)
    # group reads by sample id
    grouped_inputs = {
        key: list(val_iter) for key, val_iter in itertools.groupby(inputs, group_fn)
    }

    # If the input contains 2 files, match them to the read1 and read2 patterns
    # otherwise, assume that the input is a single file and ignore the read patterns
    sorted_grouped_reads = {}
    for key, values in grouped_inputs.items():
        if len(values) == 2:
            input1 = sorted([Path(x) for x in values if input1_pattern in str(x)])
            input2 = sorted([Path(x) for x in values if input2_pattern in str(x)])

            if len(input1) != 1:
                raise click.ClickException(
                    f"Expected an input files identified with {input1_pattern}"
                )

            if len(input2) != 1:
                raise click.ClickException(
                    f"Expected an input files identified with {input2_pattern}"
                )

            sorted_grouped_reads[key] = [input1[0], input2[0]]
        elif len(values) == 1:
            sorted_grouped_reads[key] = [Path(values[0])]
        else:
            raise ValueError(f"Unexpected number of inputs for sample {key}")

    return sorted_grouped_reads


def gz_size(filename: str) -> int:
    """Extract the size of a gzip compressed file.

    :param filename: file name
    :returns int: size of the file uncompressed (in bits)
    """
    with gzip.open(filename, "rb") as f:
        return f.seek(0, whence=2)


def log_step_start(
    step_name: str,
    input_files: Optional[List[str] | str] = None,
    output: Optional[str] = None,
    **kwargs,
) -> None:
    """Add information about the start of a pixelator step to the logs.

    :param step_name: name of the step that is starting
    :param input_files: collection of input file paths
    :param output: optional path to output
    :param **kwargs: any additional parameters that you wish to log
    :rtype: None
    """
    from pixelator import __version__

    logger.info("Start pixelator %s %s", step_name, __version__)

    if isinstance(input_files, list):
        logger.info("Input file(s) %s", ",".join(input_files))

    if isinstance(input_files, str):
        logger.info("Input file %s", input_files)

    if output is not None:
        logger.info("Output %s", output)

    if kwargs is not None:
        params = [f"{key.replace('_', '-')}={value}" for key, value in kwargs.items()]
        logger.info("Parameters:%s", ",".join(params))


def np_encoder(object: Any):
    """Encoder for JSON serialization of numpy data types."""  # noqa: D401
    if isinstance(object, np.generic):
        return object.item()


def remove_csv_whitespaces(df: pd.DataFrame) -> None:
    """Remove leading and trailing blank spaces from csv files slurped by pandas."""
    # fill NaNs as empty strings to be able to do `.str`
    df.fillna("", inplace=True)
    df.columns = df.columns.str.strip()
    for col in df.columns:
        df[col] = df[col].str.strip()


def reverse_complement(seq: str) -> str:
    """Compute the reverse complement of a DNA seq.

    :param seq: the DNA sequence
    :return: the reverse complement of the input sequence
    :rtype: str
    """
    return seq.translate(_TRTABLE)[::-1]


def sanity_check_inputs(
    input_files: Sequence[PathType] | PathType,
    allowed_extensions: Union[Sequence[str], Optional[str]] = None,
) -> None:
    """Perform basic sanity checking of input files.

    :param input_files: the files to sanity check
    :param allowed_extensions: the expected file extension of the files, e.g. 'fastq.gz'
                               or a tuple of allowed types eg. ('fastq.gz', 'fq.gz')
    :raises AssertionError: when any of validation fails
    :returns None:
    """
    input_files_: list[PathType] = (
        input_files if not isinstance(input_files, PathType) else [input_files]  # type: ignore
    )

    for input_file in input_files_:
        input_file = Path(input_file)
        logger.debug("Sanity checking %s", input_file)

        if not input_file.is_file():
            raise AssertionError(f"{input_file} is not a file")

        if input_file.stat().st_size == 0:
            raise AssertionError(f"{input_file} is an empty file")

        if not isinstance(allowed_extensions, str) and isinstance(
            allowed_extensions, collections.abc.Sequence
        ):
            if not any(str(input_file).endswith(ext) for ext in allowed_extensions):
                raise AssertionError(
                    f"{input_file} does not have any of the "
                    f"extensions {', '.join(allowed_extensions)}"
                )
        elif allowed_extensions is not None and not str(input_file).endswith(
            allowed_extensions
        ):
            raise AssertionError(
                f"{input_file} does not have the extension {allowed_extensions}"
            )


T = typing.TypeVar("T")


def single_value(xs: Union[List[T], Set[T]]) -> T:
    """Extract the first value in a List or Set if the collection has a single value.

    :param xs: a collection of values
    :returns T: the first value in the collection
    :raises AssertionError: if the collection is empty or has more than one value
    """
    if len(xs) == 0:
        raise AssertionError("Empty collection")
    if len(xs) > 1:
        raise AssertionError("More than one element in collection")
    return list(xs)[0]


def timer(func):
    """Time the different steps of a function."""

    @wraps(func)
    def wrapper(*args, **kwds):
        start_time = time.perf_counter()
        res = func(*args, **kwds)
        run_time = time.perf_counter() - start_time
        logger.info("Finished pixelator %s in %.2fs", func.__name__, run_time)
        return res

    return wrapper


def write_parameters_file(
    click_context: click.Context, output_file: Path, command_path: Optional[str] = None
) -> None:
    """Write the parameters used in for a command to a JSON file.

    :param click_context: the click context object
    :param output_file: the output file
    :param command_path: the command to use as command name
    """
    command_path_fixed = command_path or click_context.command_path
    parameters = click_context.command.params
    parameter_values = click_context.params

    param_data = {}

    for param in parameters:
        if not isinstance(param, click.core.Option):
            continue

        name = param.opts[0]
        value = parameter_values.get(str(param.name))
        if value is not None and param.type == click.Path:
            value = str(Path(value).resolve())

        param_data[name] = value

    data = {
        "cli": {
            "command": command_path_fixed,
            "options": param_data,
        }
    }

    logger.debug("Writing parameters file to %s", str(output_file))

    with open(output_file, "w") as fh:
        json.dump(data, fh, indent=4)


def _add_handlers_to_root_logger(port, log_level):
    root_logger = logging.getLogger()
    root_logger.setLevel(log_level)
    socket_handler = SocketHandler("localhost", port)
    root_logger.addHandler(socket_handler)


def _pre_multiprocessing_args(
    nbr_cores=None, logging_setup=None, context="spawn", **kwargs
):
    # If these variable are not set we will try to pick them
    # up from the click context
    current_click_context = click.get_current_context(silent=True)
    click_logging_setup = None
    click_nbr_cores = None
    if current_click_context:
        click_logging_setup = current_click_context.obj.get("LOGGER")
        click_nbr_cores = current_click_context.obj.get("CORES")

    nbr_cores = (
        nbr_cores if nbr_cores else click_nbr_cores or multiprocessing.cpu_count()
    )
    args_dict = {
        "max_workers": nbr_cores,
        "mp_context": multiprocessing.get_context(context),
    }

    if logging_setup or click_logging_setup:
        args_dict = args_dict | dict(
            initializer=_add_handlers_to_root_logger,
            initargs=(
                (logging_setup or click_logging_setup).port,
                (logging_setup or click_logging_setup).log_level,
            ),
        )
    args_dict = args_dict | kwargs
    return args_dict


def get_process_pool_executor(
    nbr_cores=None, logging_setup=None, context="spawn", **kwargs
) -> ProcessPoolExecutor:
    """Return a ProcessPool with some default settings."""
    args_dict = _pre_multiprocessing_args(nbr_cores, logging_setup, context, **kwargs)
    return ProcessPoolExecutor(**args_dict)


def get_pool_executor(
    nbr_cores=None, logging_setup=None, context="spawn", **kwargs
) -> Pool:
    """Return a Pool with some default settings."""
    args_dict = _pre_multiprocessing_args(nbr_cores, logging_setup, context, **kwargs)
    nbr_of_processes = args_dict.pop("max_workers")
    args_dict.pop("mp_context")
    return multiprocessing.get_context(context).Pool(
        processes=nbr_of_processes, **args_dict
    )


R1_REGEX = R"(.[Rr]1$)|(_[Rr]?1$)|(_[Rr]?1)(?P<suffix>_[0-9]{3})$"
R2_REGEX = R"(.[Rr]2$)|(_[Rr]?2$)|(_[Rr]?2)(?P<suffix>_[0-9]{3})$"


def get_read_sample_name(read: str) -> str:
    """Extract the sample name from a read file.

    Strip fq.gz or fastq.gz extension and remove R1/R2 suffixes.
    Supported R1 R2 identifieds are:

    _R1,_R2 | _r1, _r2 | _1, _2 | .R1, .R2 | .r1, .r2

    :param read: filename of a fastq read file
    :return str: sample name
    :raise ValueError: if the read file does not have a valid extension
    """
    # group input file by sample id and order reads by R1 and R2
    if not (read.endswith("fq.gz") or read.endswith("fastq.gz")):
        raise ValueError("Invalid file extension: expected .fq.gz or .fastq.gz")

    read_stem = Path(read).name
    read_stem = read_stem.removesuffix(get_extension(read_stem, 2)).rstrip(".")
    r1_match = re.search(R1_REGEX, read_stem)
    r2_match = re.search(R2_REGEX, read_stem)

    # Check if the r1 and r2 suffixes are "exclusive or"
    if r1_match and r2_match or (not r1_match and not r2_match):
        raise ValueError("Invalid R1/R2 suffix.")

    # We need to cast away the optional here r1 or r2 will always
    # return a match object since we checked for both being None above
    match = typing.cast(re.Match[str], r1_match or r2_match)

    # Remove the R1 or R2 suffix by using the indices returned by the match
    s, e = match.span()
    sample_name = read_stem[0:s] + read_stem[e:-1]

    if match.groupdict().get("suffix"):
        sample_name += match.group("suffix")

    return sample_name


def is_read_file(read: Path | str, read_type: Literal["r1"] | Literal["r2"]) -> bool:
    """Check if a read filename matches the specified read_type.

    Detects the presence of a common read 1 or read 2 suffix in the filename.

    :param read: filename of a fastq read file
    :param read_type: the read type to check for (r1 or r2)
    :return bool: True if the read file is a read 1 or 2 file
    :raise ValueError: if the read file does not have a valid extension
    :raise AssertionError: if the read_type is not 'r1' or 'r2'
    """
    read = Path(read).name

    if read_type not in ("r1", "r2"):
        raise AssertionError("Invalid read type: expected 'r1' or 'r2'")

    if not (read.endswith("fq.gz") or read.endswith("fastq.gz")):
        raise ValueError("Invalid file extension: expected .fq.gz or .fastq.gz")

    match: re.Match[str] | None = None
    read_stem = Path(read.removesuffix(get_extension(read, 2)).rstrip(".")).name
    if read_type == "r1":
        match = re.search(R1_REGEX, read_stem)
    elif read_type == "r2":
        match = re.search(R2_REGEX, read_stem)
    else:
        raise AssertionError(
            "Invalid read type: could not find a read suffix in filename."
        )

    if not match:
        return False

    return True


def flatten(iterable: Iterable[Iterable[Any] | Any]) -> Generator[Any, None, None]:
    """Flatten an Iterable containing items or collection of items.

    Note: only list, set, tuple are flattened, strings and bytes are yielded as is

    :param iterable: list of lists or list of sets
    :return Generator[Any, None, None]: A generator yielding the flattened items
    :yield Any: the flattened items
    """
    for item in iterable:
        if isinstance(item, (str, bytes)):
            yield item
        elif isinstance(item, (list, set, tuple)):
            yield from item
        else:
            yield item



"""Copyright © 2023 Pixelgen Technologies AB."""

import logging
import shutil
from pathlib import Path
from typing import Any, Dict, List, Optional

import pytest
from click.testing import CliRunner
from click.testing import Result as CliRunnerResult

import pixelator.cli.main as pixelator_cli

# Do not propagate to parent logger since
logger = logging.getLogger("pixelator.test_utils")


class PixelatorWorkflowContext:
    """Helper class for workflow tests.

    This class is used to run the pixelator CLI and store the results.
    It also provides access to the test configuration in yaml or as class variables
    in the test class.
    """

    def __init__(self, test_id: str, workdir: Path):
        self.test_id = test_id
        self.workdir: Path = workdir
        self.results: Dict[str, Any] = {}
        # self.config: WorkflowConfig = config
        self.commands: Dict[str, List[str]] = {}
        self.logs: Dict[str, str] = {}

    def set_logs(self, key: str, logs: str):
        self.logs[key] = logs

    def get_logs(self, key: str) -> str:
        return self.logs[key]

    def run_command(
        self, key: str, command: List[str], link_files: Optional[List[str]] = None
    ) -> CliRunnerResult:
        runner = CliRunner()

        command = [str(s) for s in command]

        with pytest.MonkeyPatch.context() as mp:
            mp.chdir(self.workdir)

            if link_files:
                self._link_files(link_files)

            logger.info(
                f"[{self.test_id}] Running pixelator command: {' '.join(command)}"
            )
            result = runner.invoke(
                pixelator_cli.main_cli, command, catch_exceptions=False
            )

            logger.info(
                f"[{self.test_id}] Command completed [exit_code: {result.exit_code}]"
            )

            if result.exit_code != 0:
                logger.warning(f"[{self.test_id}] Command failed: {result.output}")

        self.results[key] = result
        self.commands[key] = command

        pixelator_logger = logging.getLogger("pixelator")
        # Reset handlers to make sure we do not get duplicate log messages
        # This is important, otherwise the old handlers will remain active
        # and logs for all stages will be written to the stage specific log files.
        pixelator_logger.handlers = []

        return result

    def _link_files(self, files: List[str], suffix: str = ""):
        """Symlink test files into the temporary working directory.

        If no suffix is given files will be linked directly into the working directory.

        :param files: list of files to link
        :param suffix: path relative to the workflow working directory to link
            the files in.
        :raises ValueError: if a file is not absolute
        """
        for f in files:
            f_path = Path(f)
            if not f_path.is_absolute():
                raise ValueError(f"File {f} is not absolute")

            if suffix:
                target = self.workdir / suffix / f_path.name
            else:
                target = self.workdir / f_path.name

            if not target.exists():
                if f_path.is_file():
                    target.symlink_to(f_path)
                if f_path.is_dir():
                    target.symlink_to(f_path, target_is_directory=True)


@pytest.fixture(scope="class")
def use_workflow_context(request, tmp_path_factory):
    """Fixture to set up a working directory and WorkflowContext helper.

    This fixture is class scoped so that the working directory is reused across
    individual pixelator commands in a single BasePixelatorWorkflow subclass.

    The working directory is cleaned up after the test class has run unless the
    `--keep_workdirs` is passed on the commandline, which can be useful for debugging.
    """
    keep_workdirs = request.config.getoption("--keep-workdirs", default=False)
    cls = request.cls
    classname = cls.__name__
    test_id = cls.test_id

    if not hasattr(cls, "context") or cls.context is None:
        logger.info(f"Setting up working directory for {classname}")
        job_root = tmp_path_factory.mktemp(classname)
        cls.context = PixelatorWorkflowContext(test_id, job_root)

    yield

    if not keep_workdirs:
        logger.info(f"Cleaning up working directory for {classname}")
        shutil.rmtree(cls.context.workdir)



"""
Console script for pixelator (demux)

Copyright © 2022 Pixelgen Technologies AB.
"""

import sys

import click

from pixelator.common.config.panel import load_antibody_panel
from pixelator.common.utils import (
    build_barcodes_file,
    create_output_stage_dir,
    get_extension,
    get_sample_name,
    log_step_start,
    sanity_check_inputs,
    timer,
    write_parameters_file,
)
from pixelator.mpx.cli.common import (
    design_option,
    logger,
    output_option,
)
from pixelator.mpx.config import config
from pixelator.mpx.demux import demux_fastq


@click.command(
    "demux",
    short_help="demultiplex pixel data (FASTQ) to generate one file per antibody",
    options_metavar="<options>",
)
@click.argument(
    "fastq_file",
    nargs=1,
    required=True,
    type=click.Path(exists=True),
    metavar="FASTQ",
)
@click.option(
    "--mismatches",
    default=0.1,
    required=False,
    type=click.FloatRange(0.0, 0.9),
    show_default=True,
    help="The number of mismatches allowed (in percentage)",
)
@click.option(
    "--min-length",
    default=None,
    required=False,
    type=click.INT,
    show_default=False,
    help=(
        "The minimum length of the barcode that must overlap when matching.\n"
        "If you set this argument it will overrule the value from the chosen design"
    ),
)
@click.option(
    "--panel",
    required=True,
    type=str,
    help=(
        "A key of a panel file in the config,"
        " or a csv file with the antibody panel conjugations"
    ),
)
@click.option(
    "--anchored",
    default=None,
    type=click.BOOL,
    help=(
        "Enforce the barcodes to be anchored (at the end of the read).\n "
        "(default: use value determined by --design)."
    ),
)
@click.option(
    "--rev-complement",
    default=None,
    type=click.BOOL,
    help=(
        "Use the reverse complement of the barcodes sequences.\n "
        "(default: use value determined by --design)."
    ),
)
@output_option
@design_option
@click.pass_context
@timer
def demux(
    ctx,
    fastq_file,
    mismatches,
    min_length,
    panel,
    output,
    design,
    anchored,
    rev_complement,
):
    """
    Demultiplex Molecular Pixelation data (FASTQ) to generate one file per antibody
    """
    # log input parameters
    input_files = [fastq_file]
    log_step_start(
        "demux",
        input_files=input_files,
        output=output,
        mismatches=mismatches,
        panel=panel,
        min_length=min_length,
        anchored=anchored,
        rev_complement=rev_complement,
        design=design,
    )

    # some basic sanity check on the input files
    sanity_check_inputs(input_files, allowed_extensions=("fastq.gz", "fq.gz"))

    # update arguments from config file (CLI arguments have priority)
    try:
        amplicon = config.get_assay(design).get_region_by_id("amplicon")
        barcode = amplicon.get_region_by_id("bc")

        # Load default options from the design
        if min_length is None:
            min_length = barcode.min_len
        if anchored is None:
            anchored = barcode.data.get("anchored")
        if rev_complement is None:
            rev_complement = barcode.data.get("reverse_complement")

    except KeyError as err:
        raise click.ClickException(f"Parsing attribute from config file {str(err)}")

    # create output folder if it does not exist
    demux_output = create_output_stage_dir(output, "demux")

    # load marker panel
    panel = load_antibody_panel(config, panel)

    # build barcodes (fasta) on-the-fly
    barcodes = build_barcodes_file(
        panel=panel, anchored=anchored, rev_complement=rev_complement
    )

    # run cutadapt (demux mode) using parallel processing
    logger.info(f"Processing {fastq_file} with cutadapt (demux mode)")

    name = get_sample_name(fastq_file)
    extension = get_extension(fastq_file)
    output_file = demux_output / f"{name}.processed-{{name}}.{extension}"
    failed_file = demux_output / f"{name}.failed.{extension}"
    json_file = demux_output / f"{name}.report.json"

    write_parameters_file(
        ctx,
        demux_output / f"{name}.meta.json",
        command_path="pixelator single-cell-mpx demux",
    )

    results_ok = demux_fastq(
        input=str(fastq_file),
        output=str(output_file),
        failed=str(failed_file),
        report=str(json_file),
        panel=panel,
        mismatches=mismatches,
        barcodes=barcodes,
        min_length=min_length,
        cores=ctx.obj["CORES"],
        verbose=ctx.obj["VERBOSE"],
        sample_id=name,
    )

    if not results_ok:
        sys.exit(1)



"""Functions related to computing node metrics on graphs.

Copyright © 2025 Pixelgen Technologies AB.
"""

from typing import Literal

import numpy as np
import pandas as pd
import scipy as sp

from pixelator.common.marks import experimental


@experimental
def local_g(
    A: sp.sparse.csr_array,
    counts: pd.DataFrame,
    k: int = 1,
    use_weights: bool = True,
    normalize_counts: bool = False,
    W: sp.sparse.csr_array | None = None,
    method: Literal["gi", "gstari"] = "gi",
) -> pd.DataFrame:
    """Compute local G-scores for each node and marker.

    Local G([1]_) is a spatial node metric that measures the spatial association.
    The metric can for instance be used to detect hot spots for marker counts in a graph, where nodes
    that are close to each other have similar values. The metric is a Z-score that
    measures the deviation of the observed local marker expression from the expected marker
    expression under the null hypothesis of no spatial association. The sign of the score
    indicates whether the observed marker counts are higher or lower than expected
    and can therefore be used to identify hot and/or cold spots.

    The observed local marker expression for a node is computed by aggregating the weighted marker
    expression within its local neighborhood. The local G metric is largely influenced by the choice
    of edge weights and the size of the local neighborhood (`k`). The method implemented here uses
    incoming transition probabilities for a k-step random walk as edge weights. By increasing `k`,
    the local neighborhood of a node is expanded, increasing the "smoothing effect" of the metric,
    which can be useful to increase the scale of spatial association.

    Local G can also be useful for more interpretable visualization of polarized marker expression
    as it enhances spatial trends across neighborhoods in the graph, even if the marker counts within
    individual nodes are sparse.

    Note that it is important that the node ordering in the sparse adjacency matrix, and
    the counts matrix line up. If calling this method directly the caller is responsible for
    ensuring that this contract is fulfilled.

    .. [1] Bivand, R.S., Wong, D.W.S. Comparing implementations of global and
    local indicators of spatial association. TEST 27, 716–748 (2018).
    https://doi.org/10.1007/s11749-018-0599-x

    :param A: A sparse adjacency matrix representing the graph.
    :param counts: A DataFrame of marker counts for each node.
    :param k: The number of steps in the k-step random walk. Default is 1.
    :param use_weights: Whether to use weights in the computation. When turned off, all
    edge weights will be qeual to 1. Default is True.
    :param normalize_counts: Whether to normalize counts to proportions. Default is False.
    :param W: A sparse matrix of custom edge weights. This will override the automated
    computation of edge weights. `W` must have the same dimensions as A. Note that weights can
    be defined for any pair of nodes, not only the pairs represented by edges in `A`. Default is None.
    :param method: The method to use for computing local G. Must be one of 'gi' or 'gstari'.
    'gi' is the original local G metric, which does not consider self-loops, meaning that the
    local marker expression for a node is computed by aggregating the weighted expression of
    its neighbors. 'gstari' is a simplified version of local G that does consider self-loops.
    In other words, the local marker expression of a node also includes the weighted marker
    expression of the node itself. Default is 'gi'.
    :return: A DataFrame of local G-scores for each node and marker.
    :rtype: pd.DataFrame
    """
    # Check that type is one of 'gi' or 'gstari'
    if method not in ["gi", "gstari"]:
        raise ValueError("type must be one of 'gi' or 'gstari'")

    n_nodes = A.shape[0]

    marker_columns = counts.columns
    node_indices = counts.index
    counts = sp.sparse.csr_array(counts.values, dtype=np.float64)

    # Number of nodes in G must match the number of rows in counts
    if n_nodes != counts.shape[0]:
        raise ValueError(
            "The number of nodes in G must match the number of rows in counts"
        )

    # Check that k is a positive integer
    if not isinstance(k, int) or k < 1:
        raise ValueError("k must be a positive integer")

    # Convert marker counts to proportions if normalize_counts is True
    if normalize_counts:
        row_sums = counts.sum(axis=1)
        row_sums[np.abs(row_sums) < np.finfo(counts.dtype).eps] = 1
        counts = counts.multiply(1 / row_sums[:, None]).tocsr()

    if W is not None:
        # Raise ValueError if W is not a sparse matrix
        if not sp.sparse.issparse(W):
            raise ValueError("W must be a sparse matrix")
        # Raise ValueError if the dimensions of W differ from the number of nodes in G
        if W.shape[0] != n_nodes or W.shape[1] != n_nodes:
            raise ValueError("The dimensions of W must match the number of nodes in G")
    if not W:
        if use_weights:
            if method == "gstari":
                # Set diagonal of A to 1 for gstari which expects self-loops
                A = A + sp.sparse.eye(n_nodes)
            W = compute_transition_probabilities(
                A, k=k, remove_self_loops=(method == "gi")
            ).tocsr()
            # Transpose W to get incoming transition probabilities
            # This eliminates the correlation between local G and node degree
            W = W.T
        else:
            W = A
            if k > 1:
                # Expand local neighborhood using matrix powers
                W = sp.sparse.linalg.matrix_power(W, k)
                # Set all positive elements to 1
                W.data = np.ones_like(W.data)
                if method == "gstari":
                    # Set diagonal of A to 1 for gstari which expects self-loops
                    W = W + sp.sparse.eye(n_nodes)
                else:
                    # Set diagonal of W to 0 for gi to avoid self-loops
                    W.setdiag(values=0)

    # Compute lag matrix
    lag_mat = W @ counts

    if method == "gstari":
        # Compute xibar for each node and marker
        xibar_mat = np.tile(counts.mean(axis=0), n_nodes).reshape(counts.shape)
        # Compute si for each node and marker
        s_mat = np.tile(
            ((counts**2).sum(axis=0) / n_nodes) - (counts.sum(axis=0) / n_nodes) ** 2,
            n_nodes,
        ).reshape(counts.shape)

    if method == "gi":
        # Compute xibar for each node and marker, excluding i
        xibar_mat = (
            np.tile(counts.sum(axis=0), n_nodes).reshape(counts.shape) - counts
        ) / (n_nodes - 1)

        # Compute si for each node and marker, excluding i
        s_mat = (
            (
                np.tile((counts**2).sum(axis=0), n_nodes).reshape(counts.shape)
                - (counts**2)
            )
            / (n_nodes - 1)
        ) - (xibar_mat**2)

    # Calculate weights for each node. If k = 1 and use_weights=False,
    # the weights should be equal to the node degree
    weights_i = W.sum(axis=1)[:, None]
    square_weights_i = (W**2).sum(axis=1)[:, None]

    # Calculate expected value
    E_G = xibar_mat * weights_i

    # Calculate numerator G - E(G)
    numerator_mat = lag_mat - E_G

    # Calculate denomenator Var(G)
    if method == "gstari":
        var_weights = ((n_nodes * square_weights_i) - weights_i**2) / (n_nodes - 1)
        denominator_mat = np.sqrt(s_mat * var_weights)

    if method == "gi":
        var_weights = (((n_nodes - 1) * square_weights_i) - weights_i**2) / (
            n_nodes - 2
        )
        denominator_mat = np.sqrt(s_mat * var_weights)

    # Calculate Z-score
    gi_mat = numerator_mat / denominator_mat

    gi_mat[~np.isfinite(gi_mat)] = 0

    return pd.DataFrame(gi_mat, index=node_indices, columns=marker_columns)


def compute_transition_probabilities(
    A: sp.sparse.csr_array,
    k: int = 1,
    remove_self_loops: bool = False,
) -> sp.sparse.csr_array:
    """Compute transition probabilities of a graph.

    This function computes the transition probabilities for node pairs in a graph.
    Transition probabilities can for example be useful for weighting marker counts
    when computing local spatial node metrics such as local G. The transition
    probability is the probability of going from one node to another in a k-step walk,
    where `k` is the number of steps in the walk. When `k=1`, the transition probabilities
    are defined for the edges in the graph. With `k>1`, the transition probabilities can
    be positive for node pairs that are not directly connected by an edge, including
    transitions from a node to itself. In some situations, it may be desirable to ignore
    these transitions (self-loops), which can be achieved by setting `remove_self_loops=True`.
    In this case, the diagonal of the transition probability matrix is set to 0 and the
    probabilities are renormalized (row-wise) to sum to 1.

    :param A: A sparse adjacency matrix representing the graph.
    :param k: The number of steps in the random walk. Default is 1.
    :param remove_self_loops: Whether to remove self-loops from the transition probability
    matrix. Default is False.
    :return: A sparse matrix with transition probabilities.
    :rtype: sp.sparse.csr_matrix
    """
    # Check that k is a positive integer
    if not isinstance(k, int) or k < 1:
        raise ValueError("k must be a positive integer")

    W_out = A / A.sum(axis=0)[:, None]

    # Multiply W by itself k times
    if k > 1:
        W_out = sp.sparse.linalg.matrix_power(W_out, k)

    if remove_self_loops and k > 1:
        # Set diagonal of W to 0 if k > 1 for gi to avoid self-loops
        W_out.setdiag(values=0)
        # Renormalize transition probabilities to sum to 1
        row_sums = np.array(W_out.sum(axis=1)).flatten()
        inv_row_sums = np.reciprocal(row_sums, where=row_sums != 0)
        W_out = W_out.multiply(inv_row_sums[:, np.newaxis])

    return W_out



"""Marker panel management for different Molecular Pixelation assays.

Copyright © 2022 Pixelgen Technologies AB.
"""

from __future__ import annotations

import warnings
from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING, List, Optional

try:
    from typing import Self
except ImportError:
    from typing_extensions import Self


import re
from typing import TYPE_CHECKING, Optional

import pandas as pd
import ruamel.yaml as yaml

from pixelator import read_pna
from pixelator.common.config import AntibodyPanelMetadata
from pixelator.common.types import PathType
from pixelator.common.utils import logger

if TYPE_CHECKING:
    from pixelator.pna.config.config_class import PNAConfig


class PNAAntibodyPanel:
    """Class representing a Molecular Pixelation antibody panel."""

    # required columns
    _INDEX_COLUMN = "marker_id"
    _REQUIRED_COLUMNS = [
        "control",
        "sequence_1",
        "sequence_2",
        "conj_id",
    ]

    # and these should have unique values
    _UNIQUE_COLUMNS = ["sequence_1", "sequence_2", "conj_id"]

    def __init__(
        self,
        df: pd.DataFrame,
        metadata: AntibodyPanelMetadata,
        file_name: Optional[str] = None,
    ) -> None:
        """Load a panel from a dataframe and metadata.

        :param df: The dataframe containing the panel information.
        :param metadata: The metadata for the panel.
        :param file_name: The optional basename of the file from which
            the panel is loaded.

        :returns: None
        :raises AssertionError: exception if panel file is missing,
                                invalid or with incorrect format
        """
        self._filename = file_name
        self._metadata = metadata
        self._df = df

        # validate the panel
        errors = self.validate_antibody_panel(df)
        if len(errors) > 0:
            msg_str = "\n".join(errors)
            raise AssertionError(
                f"The following errors were found validating the panel: {msg_str}"
            )

    @classmethod
    def from_csv(cls, filename: PathType) -> Self:
        """Create an AntibodyPanel from a csv panel file.

        :param filename: The path to the panel file.
        :returns: The AntibodyPanel object.
        :raises AssertionError: exception if panel file is missing,
        :rtype: AntibodyPanel
        """
        panel_file = Path(filename)

        if not panel_file.is_file() or panel_file.suffix != ".csv":
            raise AssertionError(
                f"Panel file {filename} not found or has an incorrect format"
            )

        logger.debug("Creating Antibody panel from file %s", filename)

        df = cls._parse_panel(panel_file)
        metadata = cls._parse_header(panel_file)

        logger.debug("Antibody panel from file %s created", filename)

        return cls(df, metadata, file_name=panel_file.name)

    @classmethod
    def from_pxl(cls, filename: PathType) -> Self:
        """Create an AntibodyPanel from a pxl file.

        :param filename: The path to the panel file.
        :returns: The AntibodyPanel object.
        :raises AssertionError: exception if panel file is missing,
        :rtype: AntibodyPanel
        """
        pxl_file = Path(filename)

        if not pxl_file.is_file() or pxl_file.suffix != ".pxl":
            raise AssertionError(
                f"Panel file {filename} not found or has an incorrect format"
            )

        logger.debug("Creating Antibody panel from file %s", filename)

        pxl_data = read_pna(pxl_file)
        adata = pxl_data.adata()
        panel_metadata = adata.uns["panel_metadata"]
        panel_columns = panel_metadata.pop("panel_columns")
        df = adata.var[panel_columns]
        metadata = AntibodyPanelMetadata.model_validate(panel_metadata)

        logger.debug("Antibody panel from file %s created", filename)

        return cls(df, metadata, file_name=pxl_file.name)

    @property
    def name(self) -> Optional[str]:
        """The name defined in the panel metadata."""
        return self._metadata.name

    @property
    def version(self) -> Optional[str]:
        """Return the panel file version."""
        return self._metadata.version

    @property
    def description(self) -> Optional[str]:
        """Return the panel file description."""
        return self._metadata.description

    @property
    def aliases(self) -> list[str]:
        """Return the (optional) list of panel file aliases."""
        return self._metadata.aliases

    @classmethod
    def _parse_header(cls, file: Path) -> AntibodyPanelMetadata:
        """Parse front-matter YAML from a file.

        :param file: the file to parse
        :return AntibodyPanelMetadata: a pydantic model with the metadata
        """
        yaml_loader = yaml.YAML(typ="safe")

        metadata_lines = []
        with open(str(file), "r") as f:
            for line in f:
                if line.startswith("# "):
                    metadata_lines.append(line[2:])
                else:
                    break

        metadata = "".join(metadata_lines)
        raw_config = list(yaml_loader.load_all(metadata))

        if len(raw_config) == 0:
            warnings.warn(f"Expected a YAML frontmatter in {file}")
            return AntibodyPanelMetadata(version=None, name=None, description=None)

        frontmatter = raw_config[0]
        return AntibodyPanelMetadata.model_validate(frontmatter)

    @classmethod
    def _parse_panel(cls, panel_file: Path) -> pd.DataFrame:
        panel = pd.read_csv(str(panel_file), comment="#", index_col="marker_id").fillna(
            ""
        )

        panel["control"] = panel["control"].map(lambda s: s.lower() == "yes")

        return panel.copy()

    @cached_property
    def markers_control(self) -> List[str]:
        """Return a list of marker control (names)."""
        return list(self._df[self._df["control"]].index)

    @cached_property
    def markers(self) -> List[str]:
        """Return the list of unique markers in the panel."""
        return list(self._df.index.unique())

    @property
    def df(self) -> pd.DataFrame:
        """Return the panel dataframe."""
        return self._df

    @property
    def filename(self) -> Optional[str]:
        """Return the filename of the marker panel."""
        return self._filename

    @cached_property
    def size(self) -> int:
        """Return the size of the marker panel."""
        return self._df.shape[0]

    @staticmethod
    def _validate_sequences(panel_df, sequence_col):
        errors = []
        sequences = panel_df[sequence_col]
        ref_length = len(sequences.iloc[0])
        if not sequences.apply(lambda x: len(x) == ref_length).all():
            errors.append(f"All {sequence_col} values must have the same length.")

        if not sequences.str.match("^[ATCG]*$").all():
            errors.append(
                f"All {sequence_col} values must only contain ATCG characters. Offending values: "
                f"{sequences[~sequences.str.match('^[ATCG]*$')].tolist()}"
            )

        return errors

    @classmethod
    def validate_antibody_panel(cls, panel_df: pd.DataFrame) -> list[str]:
        """Validate the antibody panel dataframe.

        :param panel_df: The dataframe containing the panel information.
        :returns: A list of errors found in the panel.
        """
        errors = []

        # some basic sanity check on the panel size and columns
        if not set(cls._REQUIRED_COLUMNS).issubset(set(panel_df.columns)):
            missing_columns = set(cls._REQUIRED_COLUMNS) - set(panel_df.columns)
            errors.append(f"Panel has missing required columns: {missing_columns}")
            return errors

        if panel_df.shape[0] == 0:
            errors.append("Panel file is empty")
            return errors

        # sanity check on the unique columns
        for col in cls._UNIQUE_COLUMNS:
            if not len(panel_df[col].unique()) == len(panel_df[col]):
                errors.append(f"All values in column: {col} were not unique")

        if panel_df.index.name != cls._INDEX_COLUMN:
            errors.append(f"`{cls._INDEX_COLUMN}` is missing or is not set as index")
            return errors

        if any(panel_df.index.str.contains("_")):
            # Markers should not contain underscores since this messes
            # things up with Seurat on the R side
            errors.append(
                "The marker_id column should not contain underscores. "
                "Please use dashes instead. Offending values: "
                f"{panel_df.index[panel_df.index.str.contains('_')]}"
            )

        if panel_df["control"].dtype != bool:
            errors.append("`control` column is not boolean")

        # Check UniProt IDs format conforming to the UniProt naming convention. Empty IDs are allowed.
        if "uniprot_id" in panel_df.columns:
            # Pattern for valid UniProt IDs
            pattern = r"^[OPQ][0-9][A-Z0-9]{3}[0-9]|[A-NR-Z][0-9]([A-Z][A-Z0-9]{2}[0-9]){1,2}|$"

            def check_id(id_str):
                return all(
                    bool(re.match(pattern, id_)) for id_ in str(id_str).split(";")
                )

            bad_ids = panel_df[~panel_df["uniprot_id"].apply(check_id)]["uniprot_id"]

            if len(bad_ids) > 0:
                errors.append(
                    "Invalid UniProt IDs found."
                    "Please conform to the naming convention or remove the following IDs:"
                    f"{bad_ids.tolist()}"
                )

        errors += cls._validate_sequences(panel_df, "sequence_1")
        errors += cls._validate_sequences(panel_df, "sequence_2")

        return errors


def load_antibody_panel(config: PNAConfig, panel: PathType) -> PNAAntibodyPanel:
    """Load an antibody panel from a file or from the config file.

    :param config: the config object
    :param panel: the path to the panel file or the name of the
        panel in the config file

    :returns: the antibody panel
    :rtype: PNAAntibodyPanel
    """
    panel_str = str(panel)
    panel_from_config = config.get_panel(panel_str)

    if panel_from_config is not None:
        logger.info("Found panel in config file: %s", panel_from_config.name)
        return panel_from_config

    panel_obj = PNAAntibodyPanel.from_csv(panel)
    return panel_obj



"""Helper functions to collect data for report generation.

Copyright © 2022 Pixelgen Technologies AB.
"""

from __future__ import annotations

import logging
import typing

import numpy as np
import pandas as pd
import polars as pl

from pixelator.common.utils.simplification import simplify_line_rdp
from pixelator.mpx.pixeldataset import SIZE_DEFINITION, PixelDataset
from pixelator.mpx.report.qcreport.types import QCReportData

logger = logging.getLogger(__name__)


if typing.TYPE_CHECKING:
    from anndata import AnnData

    from pixelator.mpx.report import PixelatorWorkdir


def collect_components_umap_data(adata: AnnData) -> str:
    """Create a csv formatted string with the components umap data for the qc report.

    :param adata: an AnnData object with the umap data (obs)
    :return: a csv formatted string with umap data
    :rtype: str
    """
    empty = np.full(adata.n_obs, np.nan)

    if "X_umap" in adata.obsm:
        umap_df = pd.DataFrame(
            adata.obsm["X_umap"], columns=["umap1", "umap2"], index=adata.obs.index
        )
    else:
        umap_df = pd.DataFrame({"umap1": empty, "umap2": empty}, index=adata.obs.index)

    if "leiden" in adata.obs:
        umap_df["cluster"] = adata.obs["leiden"].to_numpy()
    else:
        umap_df["cluster"] = empty

    if "cluster_cell_class" in adata.obs:
        umap_df["cluster_cell_class"] = adata.obs["cluster_cell_class"].to_numpy()
    else:
        umap_df["cluster_cell_class"] = np.full(adata.n_obs, "unknown")

    umap_df = pd.concat(
        [
            umap_df,
            adata.obs[
                [
                    "reads",
                    "molecules",
                    "mean_b_pixels_per_a_pixel",
                    "mean_molecules_per_a_pixel",
                ]
            ],
        ],
        axis=1,
    )
    return umap_df.to_csv(index=True, index_label="component")


def collect_antibody_percentages_data(adata: AnnData) -> str:
    """Create the antibody percentages histogram data for the qc report.

    This function created a csv formatted string with the antibody name and the
    percentage of the antibody aggregated over all components.

    :param adata: an AnnData object with antibody counts percentages data
    :return: a csv formatted string with antibody percentages data
    :rtype: str
    """
    index = adata.var.index.set_names("antibody", inplace=False)
    df = pd.DataFrame(
        {"count": adata.var["antibody_count"], "percentage": adata.var["antibody_pct"]},
        index=index,
    )
    return df.to_csv()


def collect_antibody_counts_data(adata: AnnData) -> str:
    """Create the antibody counts data for the qc report.

    :param adata: an AnnData object with the antibody counts data
    :return: a csv formatted string with the antibody counts data
    :rtype: str
    """
    return adata.to_df().to_csv(index=True, index_label="component")


def collect_component_ranked_component_size_data(
    components_metrics: pd.DataFrame,
    subsample_non_cell_components: bool = False,
    subsample_epsilon: float = 1e-3,
) -> str:
    """Create data for the `cell calling` and `component size distribution` plot.

    This collects the component size and the number of antibodies per component.
    Components that pass the filters (is_filtered) are marked as selected.

    :param components_metrics: a pd.DataFrame with the components metrics
    :param subsample_non_cell_comoponents: if True, subsample non-cell components
    :param subsample_epsilon: the epsilon value for the subsampling.
        Ignored if subsample_non_cell_comoponents is False
    :return: a csv formatted string with the plotting data
    :rtype: str
    """
    component_sizes = components_metrics[SIZE_DEFINITION].to_numpy()
    df = pd.DataFrame({"component_size": component_sizes})
    df["rank"] = df.rank(method="first", ascending=False).astype(int)
    df["selected"] = components_metrics["is_filtered"].to_numpy()
    df["markers"] = components_metrics["antibodies"].to_numpy()
    df.sort_values(by="rank", inplace=True)

    if subsample_non_cell_components:
        cell_mask = df["selected"].to_numpy()
        coords = np.ascontiguousarray(df[["rank", "component_size"]].to_numpy())
        cell_idx = np.flatnonzero(cell_mask)
        other_coords_mask = ~cell_mask
        other_coords_idx = np.flatnonzero(other_coords_mask)
        other_coords = coords[other_coords_mask]

        if len(other_coords) != 0:
            simplified_coords_idx = simplify_line_rdp(
                other_coords, subsample_epsilon, return_mask=True
            )

            # Check for non empty here since zero length simplified_coords_idx
            # has shape conflicts when concatenating
            global_coords = np.concatenate(
                [cell_idx, other_coords_idx[simplified_coords_idx]]
            )
            global_coords.sort()

            df = df.iloc[global_coords]
            # These should still be sorted since we used sorted indices
            # but lets sort them again by rank just to make sure
            df.sort_values(by="rank", inplace=True)

    df.set_index("rank", inplace=True)
    return df.to_csv(index=True)


def collect_reads_per_molecule_frequency(dataset: PixelDataset) -> str:
    """Create a frequency table for the edge list "count" column.

    :param dataset: The PixelDataset object
    :return: a csv formatted string with the frequency table
    :rtype: str
    """
    freq = (
        dataset.edgelist_lazy.select(
            pl.col("count").alias("reads_per_molecule").value_counts(sort=True)
        )
        .unnest("reads_per_molecule")
        .collect()
    )

    freq = freq.with_columns(
        pl.col("reads_per_molecule"),
        frequency=pl.col("count") / pl.col("count").sum(),
    )
    pd_freq = freq.sort(by="reads_per_molecule").to_pandas()

    return pd_freq.to_csv(index=False)


def collect_report_data(workdir: PixelatorWorkdir, sample_id: str) -> QCReportData:
    """Collect the data needed to generate figures in the qc report.

    The `annotate` folder must be present in `input_path`.

    :param workdir: The PixelatorWorkdir object
    :param sample_id: The sample id
    :returns QCReportData: A QCReportData object
    :raises NotADirectoryError: If the input folder is missing the annotate folder
    :raises FileNotFoundError: If the annotate folder is missing the datasets
    """
    logger.debug("Collecting QC report data for %s in %s", sample_id, workdir.basedir)

    # parse filtered dataset
    dataset_file = workdir.filtered_dataset(sample_id)
    dataset = PixelDataset.from_file(dataset_file)
    adata = dataset.adata
    component_data = collect_components_umap_data(adata)
    antibody_percentages = collect_antibody_percentages_data(adata)
    antibody_counts = collect_antibody_counts_data(adata)

    # parse raw components metrics
    metrics_file = workdir.raw_component_metrics(sample_id)
    raw_components_metrics = pd.read_csv(metrics_file)
    ranked_component_size_data = collect_component_ranked_component_size_data(
        raw_components_metrics, subsample_non_cell_components=True
    )
    reads_per_molecule_frequency_data = collect_reads_per_molecule_frequency(dataset)

    # build the report data
    data = QCReportData(
        component_data=component_data,
        ranked_component_size=ranked_component_size_data,
        antibodies_per_cell=None,
        sequencing_saturation=None,
        antibody_percentages=antibody_percentages,
        antibody_counts=antibody_counts,
        reads_per_molecule_frequency=reads_per_molecule_frequency_data,
    )

    logger.debug("QC report data collected for %s in %s", sample_id, workdir.basedir)
    return data



"""Copyright © 2024 Pixelgen Technologies AB."""

import logging
import multiprocessing as mp
import sys
from pathlib import Path
from typing import Sequence

from cutadapt.files import FileOpener, InputPaths, OutputFiles
from cutadapt.modifiers import (
    NextseqQualityTrimmer,
    PairedEndModifier,
    PairedEndModifierWrapper,
    QualityTrimmer,
    SingleEndModifier,
)
from cutadapt.steps import PairedEndStep, SingleEndFilter, SingleEndSink, SingleEndStep
from cutadapt.utils import DummyProgress, Progress

from pixelator.pna.amplicon.build_amplicon import (
    PairedEndAmpliconBuilder,
    SingleEndAmpliconBuilder,
)
from pixelator.pna.amplicon.filters import TooManyN
from pixelator.pna.amplicon.quality import QualityProfileStep
from pixelator.pna.amplicon.report import AmpliconStatistics
from pixelator.pna.config.assay import PNAAssay
from pixelator.pna.read_processing.pipeline import AmpliconPipeline
from pixelator.pna.read_processing.runners import ParallelPipelineRunner
from pixelator.pna.utils import clean_suffixes

logger = logging.getLogger(__name__)


def amplicon_fastq(
    inputs: Sequence[Path],
    assay: PNAAssay,
    output: Path,
    mismatches: float = 0.1,
    poly_g_trimming: bool = False,
    quality_cutoff: int = 20,
    threads: int = -1,
    save_failed: bool = False,
) -> AmpliconStatistics:
    """Output a FASTQ file with amplicon sequences.

    This function will take paired-end reads and combine them into the proper amplicon
    sequences based on the assay design. The reads will be quality trimmed and filtered.

    :param inputs: The input files to process
    :param assay: The assay design to use
    :param output: The output file to write
    :param mismatches: The number of mismatches to allow
    :param poly_g_trimming: Whether to perform poly-G trimming
    :param quality_cutoff: The quality cutoff to use for quality trimming read tails
    :param threads: The number of cores to use. -1 will use all available cores
    :param save_failed: Whether to save reads that fail during amplicon combining to a separate file
    :return: A `AmpliconStatistics` instance
    """
    threads = threads if threads > 0 else mp.cpu_count()

    if len(inputs) not in [1, 2]:
        raise ValueError("Expected one or two input files, got %s" % len(inputs))

    # Open file handles for input files
    input_files = InputPaths(*(str(i) for i in inputs))

    # Open file handles for output files
    # The compression threads will be capped at 2
    write_threads = max(0, min(threads - 1, 2))
    file_opener = FileOpener(threads=write_threads)

    logging.info(
        "Opening output file writer with %s compression threads", write_threads
    )

    # Open file handles for output files
    # When writing from multiple cores the `proxied` argument will take care
    # of the necessary reordering.
    output_files = OutputFiles(
        proxied=threads > 1,
        qualities=True,
        file_opener=file_opener,
        interleaved=False,
    )

    # Open file handles for failed reads
    # TODO: Should this be configurable?
    failed_1 = Path(
        output.parent / f"{clean_suffixes(Path(inputs[0])).name}.failed.fq.zst"
    )
    if len(inputs) == 2:
        failed_2 = Path(
            output.parent / f"{clean_suffixes(Path(inputs[1])).name}.failed.fq.zst"
        )

    amplicon_failed_writer = None
    if save_failed:
        amplicon_failed_writer = output_files.open_record_writer(failed_1, failed_2)

    is_paired_end = len(inputs) == 2

    # Construct an amplicon builder class that will be used to combine the reads using the assay design

    builder_factory = (
        PairedEndAmpliconBuilder if is_paired_end else SingleEndAmpliconBuilder
    )
    builder = builder_factory(
        assay=assay, mismatches=mismatches, writer=amplicon_failed_writer
    )

    pre_steps: list[PairedEndStep | SingleEndStep] = []
    pre_modifiers: list[PairedEndModifier | SingleEndModifier] = []

    if is_paired_end:
        pre_modifiers.append(
            PairedEndModifierWrapper(
                QualityTrimmer(cutoff_front=0, cutoff_back=quality_cutoff),
                QualityTrimmer(cutoff_front=0, cutoff_back=quality_cutoff),
            )
        )

        if poly_g_trimming:
            pre_modifiers.append(
                PairedEndModifierWrapper(
                    NextseqQualityTrimmer(cutoff=20), NextseqQualityTrimmer(cutoff=20)
                )
            )
    else:
        pre_modifiers.append(
            QualityTrimmer(cutoff_front=0, cutoff_back=quality_cutoff),
        )

        if poly_g_trimming:
            pre_modifiers.append(NextseqQualityTrimmer(cutoff=20))

    sink = SingleEndSink(output_files.open_record_writer(output))

    # Construct the pipeline
    pipeline = AmpliconPipeline(
        combiner=builder,
        pre_modifiers=pre_modifiers,
        pre_steps=pre_steps,
        post_steps=[
            SingleEndFilter(predicate=TooManyN(0, assay)),
            QualityProfileStep(assay=assay),
            sink,
        ],
    )

    # Progress bar for the pipeline
    if sys.stderr.isatty():
        progress = Progress()
    else:
        progress = DummyProgress()

    n_workers = max(1, threads - write_threads)
    logging.info("Running pipeline with %s worker threads", n_workers)

    # Run the pipeline on a parallel runner
    runner = ParallelPipelineRunner(
        inpaths=input_files, n_workers=n_workers, statistics_class=AmpliconStatistics
    )

    with runner as r:
        stats = r.run(pipeline, progress, output_files)

    return stats



"""CLI for combining parquet and json files from partitioned collapse results.

Copyright © 2024 Pixelgen Technologies AB
"""

import glob

import click
import polars as pl

from pixelator.common.utils import (
    create_output_stage_dir,
    get_sample_name,
    log_step_start,
    sanity_check_inputs,
    write_parameters_file,
)
from pixelator.pna.cli.common import logger, memory_option, output_option
from pixelator.pna.collapse.independent.combine_collapse import (
    combine_independent_parquet_files,
    combine_independent_report_files,
)
from pixelator.pna.collapse.paired.combine_collapse import (
    combine_parquet_files,
    combine_report_files,
)
from pixelator.pna.collapse.report import CollapseSampleReport
from pixelator.pna.collapse.utilities import check_collapse_strategy_inputs
from pixelator.pna.utils import timer


def validate_mismatches(ctx, param, value):
    """Validate the --mismatches commandline option.

    :param ctx: The click context
    :param param: The click parameter name
    :param value: The click value
    """
    try:
        value = int(value)
        if value >= 1:
            return value
    except ValueError:
        pass

    try:
        value = float(value)
        if 0 <= value < 1:
            return value
    except ValueError:
        pass

    raise click.BadParameter("Must be an integer >= 1 or a float in range [0, 1)")


@click.command(
    "combine-collapse",
    short_help=("Combine parquet and json files from partitioned collapse results."),
    options_metavar="<options>",
)
@click.option(
    "--parquet",
    multiple=True,
    type=click.Path(exists=True),
    metavar="PARQUET",
)
@click.option(
    "--report",
    "reports",
    multiple=True,
    type=click.Path(exists=True),
    metavar="JSON",
)
@click.option(
    "--parquet-pattern",
    type=str,
    help="The pattern to match parquet files.",
)
@click.option(
    "--report-pattern",
    type=str,
    help="The pattern to match report files.",
)
@memory_option
@output_option
@click.pass_context
@timer(command_name="combine-collapse")
def combine_collapse(
    ctx, parquet, reports, output, parquet_pattern, report_pattern, memory
):
    """Collapse Molecular Pixelation data (FASTQ) by umi-upi to remove duplicates and perform error correction."""  # noqa
    # log input parameters
    log_step_start(
        "combine-collapse",
        parquet=parquet,
        report=reports,
        parquet_pattern=parquet_pattern,
        report_pattern=report_pattern,
        memory=memory,
    )
    threads = ctx.obj["CORES"] or None

    if parquet_pattern is not None:
        additional_parquet = glob.glob(parquet_pattern)
        parquet = set(parquet) if parquet else set()
        parquet.update(additional_parquet)
        parquet = list(parquet)

    if report_pattern is not None:
        additional_reports = glob.glob(report_pattern)
        reports = set(reports) if reports else set()
        reports.update(additional_reports)
        reports = list(reports)

    # some basic sanity check on the input files
    sanity_check_inputs(input_files=parquet, allowed_extensions=("parquet",))
    sanity_check_inputs(input_files=reports, allowed_extensions=("json",))

    checked_inputs = check_collapse_strategy_inputs(parquet, reports)
    collapse_output = create_output_stage_dir(output, "collapse")

    if isinstance(checked_inputs.parquet, tuple):
        umi1_files, umi2_files = checked_inputs.parquet

        # Log out all input files
        for f in umi1_files:
            logger.info(f"Using UMI1 parquet file: {str(f)}")
        for f in umi2_files:
            logger.info(f"Using UMI2 parquet file: {str(f)}")

        sample_name = get_sample_name(umi1_files[0])

        # Parameters meta.json
        write_parameters_file(
            ctx,
            collapse_output / f"{sample_name}.meta.json",
            command_path="pixelator single-cell-pna combine-collapse",
        )

        logger.info(f"Combining parquet files.")

        combined_parquet_output = collapse_output / f"{sample_name}.collapse.parquet"
        stats = combine_independent_parquet_files(
            umi1_files,
            umi2_files,
            combined_parquet_output,
            threads=threads,
            memory_limit=memory,
            temp_directory=ctx.obj.get("DUCKDB_TEMP_DIR"),
        )

        logger.info(f"Combining report files.")

        report1_files, report2_files = checked_inputs.reports
        combined_report_output = collapse_output / f"{sample_name}.report.json"

        combine_independent_report_files(
            report1_files,
            report2_files,
            sample_id=sample_name,
            stats=stats,
            output_file=combined_report_output,
        )

        return (
            umi1_files,
            umi2_files,
            output,
            ctx,
        )

    # create the output directory
    sample_name = get_sample_name(parquet[0])

    combined_output = collapse_output / f"{sample_name}.parquet"
    metrics_output = collapse_output / f"{sample_name}.report.json"

    # Parameters meta.json
    write_parameters_file(
        ctx,
        collapse_output / f"{sample_name}.meta.json",
        command_path="pixelator single-cell-pna combine-collapse",
    )

    logger.info(f"Combining parquet files.")

    # Merge parquet files
    combine_parquet_files(parquet, combined_output)

    logger.info(f"Combining report files.")

    # Merge report files
    combined_report = combine_report_files(reports)
    combined_report.add_summary_statistics(pl.scan_parquet(combined_output))

    report = CollapseSampleReport(
        product_id="single-cell-pna",
        report_type="collapse",
        sample_id=sample_name,
        **combined_report.to_dict(),
    )

    report.write_json_file(metrics_output, indent=4)



"""Model for report data returned by the single-cell analysis stage.

Copyright © 2023 Pixelgen Technologies AB.
"""

from __future__ import annotations

import pydantic

from pixelator.mpx.report.models.base import SampleReport


class PolarizationReport(pydantic.BaseModel):
    """Model for report data returned by the polarization analysis."""

    pass


class ColocalizationReport(pydantic.BaseModel):
    """Model for report data returned by the colocalization analysis."""

    pass


class AnalysisSampleReport(SampleReport):
    """Model for report data returned by the single-cell analysis stage."""

    polarization: PolarizationReport | None
    colocalization: ColocalizationReport | None



"""Model for report data returned by the single-cell preqc stage.

Copyright © 2023 Pixelgen Technologies AB.
"""

from __future__ import annotations

import json
from pathlib import Path

import pydantic

from pixelator.common.utils import get_sample_name
from pixelator.mpx.report.models.base import SampleReport


class PreQCSampleReport(SampleReport):
    """Model for data returned by the preqc stage."""

    total_read_count: int = pydantic.Field(
        ..., description="The total number of input reads in the preqc stage."
    )

    passed_filter_read_count: int = pydantic.Field(
        ...,
        description="The number of reads that passed the filters in the preqc stage.",
    )

    low_quality_read_count: int = pydantic.Field(
        ..., description="The number of low quality reads."
    )

    too_many_n_read_count: int = pydantic.Field(
        ..., description="The number of reads discarded because of too many Ns."
    )

    too_short_read_count: int = pydantic.Field(
        ..., description="The number of reads discarded because they are too short."
    )

    too_long_read_count: int = pydantic.Field(
        ..., description="The number of reads discarded because they are too long."
    )

    @pydantic.computed_field(  # type: ignore
        return_type=float,
        description="The fraction of reads that was discarded in this stage.",
    )
    @property
    def discarded(self) -> float:  # noqa: D102
        return 1 - (self.passed_filter_read_count / self.total_read_count)

    @classmethod
    def from_json(cls, p: Path) -> PreQCSampleReport:
        """Initialize an :class:`PreQCSampleReport` from a report file."""
        sample_name = get_sample_name(p)

        with open(p) as fp:
            json_data = json.load(fp)

        filtering_results = json_data["filtering_result"]

        data = {
            "total_read_count": json_data["summary"]["before_filtering"]["total_reads"],
            "passed_filter_read_count": filtering_results["passed_filter_reads"],
            "low_quality_read_count": filtering_results["low_quality_reads"],
            "too_many_n_read_count": filtering_results["too_many_N_reads"],
            "too_short_read_count": filtering_results["too_short_reads"],
            "too_long_read_count": filtering_results["too_long_reads"],
        }

        return cls(sample_id=sample_name, **data)



"""Entrypoint for demultiplexing functions.

Copyright © 2024 Pixelgen Technologies AB
"""

from .process import (
    correct_marker_barcodes,
    demux_barcode_groups,
    finalize_batched_groups,
)

__all__ = ["correct_marker_barcodes", "demux_barcode_groups", "finalize_batched_groups"]



"""Model for report data returned by the single-cell demux stage.

Copyright © 2023 Pixelgen Technologies AB.
"""

from __future__ import annotations

import json
from pathlib import Path

import pydantic

from pixelator.common.utils import get_sample_name
from pixelator.mpx.report.models.base import SampleReport


class DemuxSampleReport(SampleReport):
    """Model for data returned by the demux stage."""

    input_read_count: int = pydantic.Field(
        ...,
        description="The total number of input reads in the demux stage.",
    )

    output_read_count: int = pydantic.Field(
        ...,
        description="The total number of reads that passed the demux stage.",
    )

    per_antibody_read_counts: dict[str, int] = pydantic.Field(
        ...,
        description="The number of reads per antibody after demultiplexing.",
    )

    @pydantic.computed_field(  # type: ignore
        return_type=dict[str, float],
        description="The fraction of reads per antibody after demultiplexing.",
    )
    @property
    def per_antibody_read_count_fractions(self) -> dict[str, float]:  # noqa: D102
        return {
            antibody: count / self.output_read_count
            for antibody, count in self.per_antibody_read_counts.items()
        }

    @pydantic.computed_field(  # type: ignore
        return_type=int,
        description="Number of reads without a recognized antibody barcode.",
    )
    @property
    def unrecognised_antibody_read_count(self) -> int:  # noqa: D102
        return self.input_read_count - self.output_read_count

    @pydantic.computed_field(  # type: ignore
        return_type=int,
        description="Fraction of reads without a recognized antibody barcode.",
    )
    @property
    def fraction_unrecognised_antibody_reads(self) -> float:  # noqa: D102
        return self.unrecognised_antibody_read_count / self.input_read_count

    @classmethod
    def from_json(cls, p: Path) -> DemuxSampleReport:
        """Initialize an :class:`DemuxSampleReport` from a cutadapt report file."""
        sample_name = get_sample_name(p)

        with open(p) as fp:
            json_data = json.load(fp)

        input_read_count = json_data["read_counts"]["input"]
        output_read_count = json_data["read_counts"]["read1_with_adapter"]

        per_antibody_read_counts = dict()

        for antibody_json in sorted(
            json_data["adapters_read1"], key=lambda x: x["name"]
        ):
            antibody_name = antibody_json["name"]
            per_antibody_read_counts[antibody_name] = antibody_json["total_matches"]

        return cls(
            sample_id=sample_name,
            input_read_count=input_read_count,
            output_read_count=output_read_count,
            per_antibody_read_counts=per_antibody_read_counts,
        )



"""Collection of functions for the concatenation of raw fastq reads.

Copyright © 2022 Pixelgen Technologies AB.
"""

import logging

# List is used as type hint in comment
from typing import BinaryIO, Sequence, Tuple  # noqa: F401

import pyfastx
from xopen import xopen

from pixelator.common.config import Region
from pixelator.common.types import PathType
from pixelator.common.utils import reverse_complement
from pixelator.mpx.amplicon.statistics import SequenceQualityStatsCollector
from pixelator.mpx.config import config
from pixelator.mpx.report.models import AmpliconSampleReport

logger = logging.getLogger(__name__)

ILLMN_QUAL_0 = "!"

FastxReadTuple = Tuple[str, str, str]


def trim_amplicon(record1: FastxReadTuple, amplicon: Region) -> FastxReadTuple:
    """Trim a single-end fastq record to match the amplicon length.

    :param record1: the forward fastq record
    :param amplicon: the amplicon region from the assay specification
    :returns: a tuple with the sequence id, sequence and quality
    :rtype: FastxReadTuple
    """
    n1, s1, q1 = record1
    s1_len = len(s1)

    # Full amplicon with Ns placeholders for random sequences
    min_amplicon_len, max_amplicon_len = amplicon.get_len()

    if s1_len <= min_amplicon_len:
        return record1

    amplicon_seq = s1[:min_amplicon_len]
    amplicon_qual = q1[:min_amplicon_len]

    return n1, amplicon_seq, amplicon_qual


def generate_amplicon(
    record1: FastxReadTuple, record2: FastxReadTuple, amplicon: Region
) -> FastxReadTuple:
    """Generate a read from paired end fastq records.

    :param record1: a tuple with name, sequence and quality of the forward reads
    :param record2: a tuple with name, sequence and quality of the reverse read
    :param amplicon: the amplicon region from the assay specification
    :return: a tuple with the name, sequence and quality
    :rtype: FastxReadTuple
    :raises ValueError: if the headers of the two records are different
    """
    n1, s1, q1 = record1
    n2, s2, q2 = record2

    if n1 != n2:
        msg = "Found different headers when parsing fastq files"
        raise ValueError(msg)

    s1_len = len(s1)
    s2_len = len(s2)

    # Full amplicon with N placeholders for random sequences
    ref_amplicon_seq = amplicon.get_sequence()
    min_amplicon_len, max_amplicon_len = amplicon.get_len()
    ref_amplicon_qual = "!" * len(ref_amplicon_seq)

    # R1 already longer than amplicon -> just trim similar to single-end
    if s1_len >= min_amplicon_len:
        amplicon_seq = s1[:min_amplicon_len]
        amplicon_qual = q1[:min_amplicon_len]

        return n1, amplicon_seq, amplicon_qual

    # R2 already longer than amplicon -> just trim similar to single-end
    if s2_len >= min_amplicon_len:
        amplicon_seq = reverse_complement(s2)[:min_amplicon_len]
        amplicon_qual = q1[::-1][:min_amplicon_len]

        return n1, amplicon_seq, amplicon_qual

    # The amplicon sequence and quality strings
    # are overwritten with R1 from the start and R2 from the end.
    amplicon_seq = s1 + ref_amplicon_seq[s1_len:]
    amplicon_qual = q1 + ref_amplicon_qual[s1_len:]

    amplicon_seq = amplicon_seq[:-s2_len] + reverse_complement(s2)
    amplicon_qual = amplicon_qual[:-s2_len] + q2[::-1]

    return n1, amplicon_seq, amplicon_qual


def write_record(f: BinaryIO, header: str, sequence: str, quality: str) -> None:
    """Write a fastq record to a file.

    :param f: the file to write to
    :param header: the header of the record
    :param sequence: the sequence of the record
    :param quality: the quality of the record
    :rtype: None
    """
    # Do not generate intermediate strings here to avoid unneeded copies
    f.write(b"@")
    f.write(header.encode("utf-8"))
    f.write(b"\n")
    f.write(sequence.encode("utf-8"))  # type: ignore
    f.write(b"\n+\n")
    f.write(quality.encode("utf-8"))  # type: ignore
    f.write(b"\n")


def amplicon_fastq(
    inputs: Sequence[PathType],
    design: str,
    metrics: PathType,
    sample_id: str,
    output: PathType,
) -> None:
    """Build MPX amplicons and save them to fastq files.

    The amplicon building process works without trying to overlap PE reads at the
    moment, even if PE reads would overlap. This function will iterate through the
    contents of the MPX fastq files and generate a new one based on the designed
    amplicon.

    :param inputs: a list of path to the fastq reads
    :param design: the design used in the config file
    :param metrics: the path to the json metrics file
    :param sample_id: the sample id
    :param output: the path to the output file (processed)
    :returns: None
    :rtype: None
    :raises RuntimeError: raises an exception
    """
    logger.debug("Using design %s", design)
    assay = config.get_assay(design)

    if assay is None:
        raise RuntimeError(f"Unknown design {design}")

    if not 0 < len(inputs) <= 2:
        raise RuntimeError("Invalid number of input files, expected 1 or 2 files")

    mode = "single-end" if len(inputs) == 1 else "paired-end"
    stats = SequenceQualityStatsCollector(design)

    amplicon = assay.get_region_by_id("amplicon")
    if amplicon is None:
        raise RuntimeError("Design does not have a region with id: amplicon")

    # Single end mode
    if mode == "single-end":
        logger.debug("Building amplicon of %s to %s", inputs[0], output)

        with xopen(output, "wb") as f:
            for record in pyfastx.Fastq(str(inputs[0]), build_index=False):
                name, new_seq, new_qual = trim_amplicon(record, amplicon)
                write_record(f, name, new_seq, new_qual)
                stats.update(new_qual)

    if mode == "paired-end":
        logger.debug(
            "Building amplicon of %s and %s to %s", inputs[0], inputs[1], output
        )

        with xopen(output, "wb") as f:
            for idx, (record1, record2) in enumerate(
                zip(
                    pyfastx.Fastq(str(inputs[0]), build_index=False),
                    pyfastx.Fastq(str(inputs[1]), build_index=False),
                )
            ):
                if idx % 100000 == 0:
                    logger.debug(
                        "Generating amplicon for %s: %s reads processed",
                        str(output),
                        str(idx),
                    )

                name, new_seq, new_qual = generate_amplicon(record1, record2, amplicon)
                write_record(f, name, new_seq, new_qual)
                stats.update(new_qual)

    logger.info(
        "Generating amplicon for %s: %s reads processed", str(output), stats.read_count
    )
    # add metrics to JSON file
    avg_stats = stats.stats

    report = AmpliconSampleReport(sample_id=sample_id, **avg_stats.asdict())
    report.write_json_file(metrics, indent=4)

    if mode == "single-end":
        logger.debug("Finished building amplicon of %s to %s", inputs[0], output)
    else:
        logger.debug(
            "Finished building amplicon of %s and %s to %s",
            inputs[0],
            inputs[1],
            output,
        )



"""Annotation tests for the graph step.

Copyright © 2023 Pixelgen Technologies AB.
"""

import logging

import pytest

from .base import BaseWorkflowTestMixin

logger = logging.getLogger(__name__)


class BaseGraphTestsMixin(BaseWorkflowTestMixin):
    """Base class for graph command tests.

    Test cases (defined in this class or in subclasses)
    that depend on the output should be marked with:
    ```
    @pytest.mark.dependency(depends=["test_graph_run"])
    ```
    """

    __stage_key__ = "graph"

    @pytest.mark.dependency(name="test_graph_run", depends=["test_collapse_run"])
    def test_graph_run(self):
        """Run the graph command."""
        params = self.__get_parameters()
        verbose = self.__get_options("common").get("verbose")

        input_files = list(
            map(
                lambda f: str(f),
                (self.workdir / "collapse").glob("*.collapsed.parquet"),
            )
        )

        command = [
            "--log-file",
            "graph-pixelator.log",
            "single-cell-mpx",
            "graph",
            "--output",
            ".",
        ]

        if verbose:
            command = ["--verbose"] + command

        if params:
            command += params

        command += input_files

        self.context.run_command("graph", command, input_files)
        assert self.__this_result.exit_code == 0

    @pytest.mark.dependency(depends=["test_graph_run"])
    def test_graph_edgelist_exists(self):
        """Check that the edgelist exists."""
        edge_files = (self.workdir / "graph").glob("*.edgelist.parquet")
        for f in edge_files:
            assert f.is_file()

    @pytest.mark.dependency(depends=["test_graph_run"])
    def test_graph_report_exists(self):
        """Check that the json report exists."""
        json_files = (self.workdir / "graph").glob("*.report.json")
        for f in json_files:
            assert f.is_file()



"""Analysis classes for the PNA data.

Copyright © 2024 Pixelgen Technologies AB.
"""

import logging
import tempfile
from functools import partial
from pathlib import Path

import numpy as np
import pandas as pd
import polars as pl

from pixelator.pna.analysis.proximity import jcs_with_permute_stats
from pixelator.pna.analysis_engine import PerComponentTask
from pixelator.pna.graph import PNAGraph
from pixelator.pna.pixeldataset import PNAPixelDataset
from pixelator.pna.pixeldataset.io import PixelFileWriter, PxlFile

logger = logging.getLogger(__name__)


class ProximityAnalysis(PerComponentTask):
    """Run proximity analysis on each component."""

    TASK_NAME = "proximity"

    def __init__(
        self,
        n_permutations: int = 100,
        min_marker_count: int = 10,
    ) -> None:
        """Initialize a ProximityAnalysis instance.

        :param n_permutations: the number of permutations to use for the method (only used with methods
                               that use permutations).
        """
        self.method = "join_count_statistics"
        self._proxmimity_function = partial(
            jcs_with_permute_stats,
            n_permutations=n_permutations,
            min_marker_count=min_marker_count,
        )

    def run_on_component_edgelist(
        self, component_df: pl.LazyFrame, component_id: str
    ) -> pd.DataFrame:
        """Run proximity analysis on a single component.

        :param component: a Graph for a component to run the analysis on.
        :param component_id: the id of the component.
        :return: a pandas DataFrame containing average_node degrees per type counts.
        """
        result = self._proxmimity_function(component_df.collect())
        result["component"] = component_id
        return result

    def add_to_pixel_file(self, data: pd.DataFrame, pxl_file_target: PxlFile) -> None:
        """Add proximity data for all components to the pixeldataset.

        :param data: a pandas DataFrame containing proximity data for all components.
        :param pxl_dataset: the PixelDataset to add the data to.
        """
        logger.debug("Adding proximity data to PixelDataset")
        with tempfile.NamedTemporaryFile(suffix=".parquet") as tmp_file:
            tmp_file = Path(tmp_file.name)  # type: ignore
            data.to_parquet(tmp_file, index=False)
            with PixelFileWriter(pxl_file_target.path) as writer:
                writer.write_proximity(tmp_file)
        logger.debug("Proximity data added to PixelDataset")

    def run_on_component_graph(
        self, component: PNAGraph, component_id: str
    ) -> pd.DataFrame:
        """Run proximity analysis on a single component."""
        raise NotImplementedError

    def parameters(self) -> dict:
        """Return the parameters of the `PerComponentAnalysis`.

        This is used e.g. to store the metadata the parameters of the analysis
        in the run metadata.
        """
        data = super().parameters()
        data["method"] = self.method
        return data



"""Contains constants used as part of plotting modules.

Copyright © 2023 Pixelgen Technologies AB.
"""


class Color:
    """Class for color constants."""

    LIGHTGREY = "#bebebe"
    ORANGERED2 = "#ee4000"
    SKYBLUE3 = "#6ca6cd"
    NETWORKX_NODE_COLOR = "#1f78b4"
    JETSET = [
        "#00007F",  # dark blue
        "#0000FF",  # blue
        "#007FFF",  # azure
        "#00FFFF",  # cyan
        "#7FFF7F",  # light green
        "#FFFF00",  # yellow
        "#FF7F00",  # orange
        "#FF0000",  # red
        "#7F0000",  # dark red
    ]



"""Functions related to perform community detection on the pixelator graph step.

Copyright © 2022 Pixelgen Technologies AB.
"""

import logging
from copy import copy
from pathlib import Path
from typing import Tuple

import networkx as nx
import numpy as np
import pandas as pd
import polars as pl
from graspologic_native import leiden

from pixelator.mpx.graph.constants import (
    LEIDEN_RESOLUTION,
    MIN_PIXELS_TO_REFINE,
)
from pixelator.mpx.graph.utils import (
    edgelist_metrics,
    map_upis_to_components,
    split_remaining_and_removed_edgelist,
)
from pixelator.mpx.report.models.graph import GraphSampleReport

logger = logging.getLogger(__name__)


def connect_components(
    input: str,
    output: str,
    sample_name: str,
    metrics_file: str,
    multiplet_recovery: bool,
    max_refinement_recursion_depth: int = 5,
    max_edges_to_split: int = 5,
    min_count: int = 2,
) -> None:
    """Retrieve all connected components from an edgelist.

    This function takes as input an edge list in `parquet` format that has
    been generated with `pixelator collapse`. The function filters the
    edge list by count (`min_count`) and then adds a column to the edge list
    with the respective connected components ids obtained from the graph. The
    column is named "component". The edge list is then processed to recover
    big components (technical multiplets) into smaller components if only if
    `multiplet_recovery` is True. The recovery is done using community
    detection to detect and remove problematic edges using the Leiden [1]_ community
    detection algorithm. Information about the recovered components is written to
    a CSV file (`<sample_id>.components_recovered.csv`) and the filtered edge list
    is written to a parquet file (`<sample_id>`.edgelist.parquet) alongside the
    discarded edges (`<sample_id>`.discarded_edgelist.parquet).

    The following files are generated:

    - edge list (parquet) after multiplets recovery (if any)
    - metrics (json) with information


    .. [1] Traag, V.A., Waltman, L. & van Eck, N.J. From Louvain to Leiden: guaranteeing
        well-connected communities. Sci Rep 9, 5233 (2019).
        https://doi.org/10.1038/s4q:598-019-41695-z

    :param input: the path to the edge list dataframe (parquet)
    :param output: the path to the output folder
    :param sample_name: the prefix to prepend to the files (sample name)
    :param metrics_file: the path to a JSON file to write metrics
    :param multiplet_recovery: set to true to activate multiplet recovery
    :param max_refinement_recursion_depth: The number of times a component can be broken down into
                             smaller components during the recovery process.
    :param max_edges_to_split: The maximum number of edges between the product components
                                when splitting during multiplet recovery.
    :param min_count: the minimum number of counts (molecules) an edge must have
    :returns: None
    :rtype: None
    :raises RuntimeError: if the edge list is empty after filtering
    """
    logger.debug("Parsing edge list %s", input)

    # load data (edge list in data frame format)
    edgelist = pl.scan_parquet(input, low_memory=True)

    nbr_of_rows = edgelist.select(pl.count()).collect()[0, 0]
    # filter data by count
    if min_count > 1:
        logger.debug(
            "Filtering edge list with %i rows using %i as minimum count",
            nbr_of_rows,
            min_count,
        )
        edgelist = edgelist.filter(pl.col("count") >= min_count)
        nbr_of_rows = edgelist.select(pl.count()).collect()[0, 0]
        logger.debug("Filtered edge list has %i elements", nbr_of_rows)

    if nbr_of_rows == 0:
        raise RuntimeError(
            f"The edge list has 0 elements after filtering by %{min_count}"
        )

    # check if the are problematic edges (same upib and upia)
    problematic_edges = set(
        np.intersect1d(
            edgelist.select("upib").unique().collect().to_numpy(),
            edgelist.select("upia").unique().collect().to_numpy(),
        )
    )
    if len(problematic_edges) > 0:
        logger.warning(
            "The edge list has %i intersecting UPIA and UPIB, these will be removed",
            len(problematic_edges),
        )
        edgelist_no_prob = edgelist.filter(
            (~pl.col("upib").is_in(problematic_edges))
            & (~pl.col("upia").is_in(problematic_edges))
        )
    else:
        edgelist_no_prob = edgelist

    nbr_of_rows = edgelist_no_prob.select(pl.count()).collect()[0, 0]
    if nbr_of_rows == 0:
        raise RuntimeError(
            "The edge list has 0 elements after removing problematic edges"
        )

    edges = (
        edgelist_no_prob.select(["upia", "upib"])
        .group_by(["upia", "upib"])
        .len()
        .sort(["upia", "upib"])  # sort to make sure the graph is the same
        .collect()
        .to_pandas()
    )
    graph = nx.from_pandas_edgelist(edges, source="upia", target="upib")

    node_component_map = pd.Series(index=graph.nodes())
    for i, cc in enumerate(nx.connected_components(graph)):
        node_component_map[list(cc)] = i
    del graph

    # get raw metrics before multiplets recovery
    logger.debug("Calculating raw edgelist metrics")

    if multiplet_recovery:
        recovered_node_component_map, node_depth_map = recover_technical_multiplets(
            edgelist=edges,
            node_component_map=node_component_map.astype(np.int64),
            max_refinement_recursion_depth=max_refinement_recursion_depth,
            max_edges_to_split=max_edges_to_split,
        )
    else:
        recovered_node_component_map = node_component_map
        node_depth_map = pd.Series(index=node_component_map.index, data=0)

    del edges

    # assign component column to edge list
    edgelist_with_component_info = map_upis_to_components(
        edgelist=edgelist,
        node_component_map=recovered_node_component_map.astype(np.int64),
        node_depth_map=node_depth_map.astype(np.int64),
    )
    remaining_edgelist, removed_edgelist = split_remaining_and_removed_edgelist(
        edgelist_with_component_info
    )

    # save the edge list (discarded)
    logger.debug("Save discarded edge list")
    removed_edgelist_df = removed_edgelist.collect()
    removed_edgelist_df.write_parquet(
        Path(output) / f"{sample_name}.discarded_edgelist.parquet"
    )

    # save the edge list (recovered)
    logger.debug("Save the edgelist")
    graph_output_edgelist = remaining_edgelist.collect()
    graph_output_edgelist.write_parquet(
        Path(output) / f"{sample_name}.edgelist.parquet",
        compression="zstd",
    )

    logger.debug("Generate graph report")
    result_metrics = edgelist_metrics(graph_output_edgelist)

    result_metrics["edges_with_colliding_upi_count"] = int(
        (removed_edgelist_df["depth"] == 0).sum()
    )
    result_metrics["edges_removed_in_multiplet_recovery_first_iteration"] = int(
        (removed_edgelist_df["depth"] == 1).sum()
    )
    result_metrics["edges_removed_in_multiplet_recovery_refinement"] = int(
        (removed_edgelist_df["depth"] > 1).sum()
    )
    result_metrics["fraction_edges_removed_in_refinement"] = float(
        (removed_edgelist_df["depth"] > 1).sum() / max(len(removed_edgelist_df), 1)
    )

    del graph_output_edgelist
    del removed_edgelist_df

    report = GraphSampleReport(
        sample_id=sample_name,
        **result_metrics,
    )
    report.write_json_file(Path(metrics_file), indent=4)


def merge_strongly_connected_communities(
    edgelist: pd.DataFrame,
    node_community_dict: dict,
    n_edges: int | None = None,
) -> Tuple[pd.DataFrame, pd.Series]:
    """Merge strongly connected communities in an edge list.

    This function takes an edge list and a dictionary with the community mapping
    for each node. It then computes the number of edges between communities and
    if they are higher than the given n_edges. It assigns the same community id
    to the nodes in the connected strongly connected communities. If `n_edges` is
    None, the split communities are not considered for merging.

    :param edgelist: The edge list to process
    :param node_community_dict: A dictionary with the community mapping for each node
    :param n_edges: The threshold for the number of edges to be found between communities to merge or None to avoid merging
    :returns: A tuple with the modified edge list and the updated community mapping
    """
    community_serie = pd.Series(node_community_dict)
    edgelist["upia_community"] = community_serie[edgelist["upia"]].values
    edgelist["upib_community"] = community_serie[edgelist["upib"]].values

    if n_edges is None or (community_serie.nunique() == 1):
        return edgelist, community_serie

    edge_counts = (
        edgelist.groupby(["upia_community", "upib_community"])["upia"]
        .count()
        .unstack(fill_value=0)
    )
    edge_counts = edge_counts.add(edge_counts.T, fill_value=0)
    cross_community_edges = edge_counts.where(
        np.tril(np.ones(edge_counts.shape), k=-1).astype(bool)
    ).stack()
    connected_communities = cross_community_edges[cross_community_edges > n_edges].index
    communities_graph = nx.from_edgelist(connected_communities)
    for cc in nx.connected_components(communities_graph):
        new_tag = min(cc)
        community_serie[community_serie.isin(cc)] = new_tag
        edgelist.loc[edgelist["upia_community"].isin(cc), "upia_community"] = new_tag
        edgelist.loc[edgelist["upib_community"].isin(cc), "upib_community"] = new_tag
    return edgelist, community_serie


def recover_technical_multiplets(
    edgelist: pd.DataFrame,
    node_component_map: pd.Series,
    max_refinement_recursion_depth: int = 5,
    max_edges_to_split: int = 5,
) -> Tuple[pd.Series, pd.Series]:
    """Perform component recovery by deleting spurious edges.

    The molecular pixelation assay may under some conditions introduce spurious
    edges. This creates components that are made up of two or more internally
    well-connected graphs, that are connected to each other by a low number of edges
    (typically in the single digits). We call these components technical multiplets.
    This method will attempt to break up these components by performing community
    detection on the whole graph using the Leiden algorithm [1]_.

    The community detection algorithm will attempt to find communities (i.e. subgraphs)
    that are internally well-connected. Edges between these communities are computed
    and consequently removed from the edge list to create new components that are
    internally well connected, and that should represent real cells.


    .. [1] Traag, V.A., Waltman, L. & van Eck, N.J. From Louvain to Leiden: guaranteeing
        well-connected communities. Sci Rep 9, 5233 (2019).
        https://doi.org/10.1038/s4q:598-019-41695-z

    :param edgelist: The edge list used to create the graph
    :param node_component_map: A series with the component mapping for each
                            node where the index is the upi (i.e. node name)
                            and the value is the component id.
    :param max_refinement_recursion_depth: The number of times a component can be broken down
                            into smaller components during the recovery process.
    :param max_edges_to_split: The maximum number of edges between the product components
                            when splitting during multiplet recovery.
    :return: A tuple with the updated node component map and the iteration depth at which each
             node is re-assigned to a component.
    :rtype: Tuple[pd.Series, pd.Series]
    """
    logger.debug(
        "Starting multiplets recovery in edge list with %i rows",
        edgelist.shape[0],
    )

    def id_generator(start=0):
        next_id = start
        while True:
            yield next_id
            next_id += 1

    id_gen = id_generator(node_component_map.max() + 1)
    comp_sizes = node_component_map.groupby(node_component_map).count()

    n_edges_to_remove = 0
    to_be_refined_next = comp_sizes[comp_sizes > MIN_PIXELS_TO_REFINE].index
    node_depth_map = pd.Series(index=node_component_map.index, data=0)
    for depth in range(max_refinement_recursion_depth):
        edgelist["component_a"] = node_component_map[edgelist["upia"]].values
        edgelist["component_b"] = node_component_map[edgelist["upib"]].values
        component_groups = edgelist.groupby("component_a")
        to_be_refined = copy(to_be_refined_next)
        to_be_refined_next = []
        for component, component_edgelist in component_groups:
            if component not in to_be_refined:
                continue

            component_edgelist = component_edgelist[
                component_edgelist["component_b"] == component
            ].sort_values(["upia", "upib"])

            edgelist_tuple = list(
                map(tuple, np.array(component_edgelist[["upia", "upib", "len"]]))
            )

            # run the leiden algorithm to get the communities
            _, community_dict = leiden(
                edgelist_tuple,
                resolution=LEIDEN_RESOLUTION
                if depth > 0
                else 1.0,  # Higher initial resolution to break up the mega-cluster
                seed=42,
                trials=5,
                # These parameters are used to sync up the native implementation with
                # the python implementation we originally used.
                iterations=1,
                randomness=0.001,
                use_modularity=True,
                starting_communities=None,
            )

            component_edgelist, community_serie = merge_strongly_connected_communities(
                component_edgelist,
                community_dict,
                n_edges=max_edges_to_split if depth > 0 else None,
            )

            if community_serie.nunique() == 1:
                continue
            n_edges_to_remove += (
                component_edgelist["upia_community"]
                != component_edgelist["upib_community"]
            ).sum()
            community_size_map = community_serie.groupby(community_serie).count()
            node_depth_map[community_serie.index] = depth + 1

            if (community_size_map > MIN_PIXELS_TO_REFINE).sum() > 1:
                further_refinement = True
            else:
                further_refinement = False

            for new_community in community_size_map.index:
                new_id = next(id_gen)
                node_component_map[
                    community_serie[community_serie == new_community].index
                ] = new_id
                if (
                    further_refinement
                    and community_size_map[new_community] > MIN_PIXELS_TO_REFINE
                ):
                    to_be_refined_next.append(new_id)

    logger.info(
        "Obtained %i components after removing %i edges",
        node_component_map.nunique(),
        n_edges_to_remove,
    )
    return node_component_map, node_depth_map



"""Model for report data returned by the annotate stage.

Copyright © 2023 Pixelgen Technologies AB.
"""

from __future__ import annotations

import textwrap
from typing import Optional

import pydantic

from pixelator.common.report.models.summary_statistics import SummaryStatistics
from pixelator.mpx.report.models.base import SampleReport


class AnnotateSampleReport(SampleReport):
    """Model for report data returned by the annotate stage."""

    fraction_molecules_in_largest_component: float = pydantic.Field(
        description="The fraction of molecules in the largest component.",
    )

    fraction_pixels_in_largest_component: float = pydantic.Field(
        description="The fraction of pixels (A and B pixels) in the largest component.",
    )

    fraction_potential_doublets: float | None = pydantic.Field(
        description=(
            "The fraction of components that appear to consist of multiple "
            "parts by the community detection algorithm."
        ),
    )

    n_edges_to_split_potential_doublets: int | None = pydantic.Field(
        description=(
            "The total number of edges that need to be removed to split the "
            "potential doublets into their sub-communities."
        ),
    )

    # ------------------------------------------------------------------------------- #
    #   Annotate metrics
    # ------------------------------------------------------------------------------- #
    input_cell_count: int = pydantic.Field(
        description="The total number of cell components in the input before filtering.",
    )

    input_read_count: int = pydantic.Field(
        description="The total number of reads in the input before filtering.",
    )

    # cells_filtered: int
    cell_count: int = pydantic.Field(
        description="The total number of cells after filtering for component size."
    )

    marker_count: int = pydantic.Field(
        description="The total number of detected antibodies after filtering for component size.",
    )

    total_marker_count: int = pydantic.Field(
        description="The total number of antibodies defined by the panel.",
    )

    molecule_count: int = pydantic.Field(
        description="The total number of unique molecules in cell or aggregate components.",
    )

    a_pixel_count: int = pydantic.Field(
        description="The number of unique A-pixels in the graph.",
    )

    b_pixel_count: int = pydantic.Field(
        description="The number of unique B-pixels in the graph.",
    )

    @pydantic.computed_field(
        return_type=int,
        description="The total number of unique pixels in the graph.",
    )
    def pixel_count(self) -> int:  # noqa: D102
        return self.a_pixel_count + self.b_pixel_count

    read_count: int = pydantic.Field(
        description="The total number of reads for all unique molecules in cell or aggregate components.",
    )

    molecule_count_per_cell_stats: SummaryStatistics = pydantic.Field(
        description="Summary statistics for the number of molecules per cell component.",
    )

    read_count_per_cell_stats: SummaryStatistics = pydantic.Field(
        description="Summary statistics for the number of reads per cell component.",
    )

    a_pixel_count_per_cell_stats: SummaryStatistics = pydantic.Field(
        description="Summary statistics for the number of A-pixels per cell component.",
    )

    b_pixel_count_per_cell_stats: SummaryStatistics = pydantic.Field(
        description="Summary statistics for the number of B-pixels per cell component.",
    )

    marker_count_per_cell_stats: SummaryStatistics = pydantic.Field(
        description="Summary statistics for the number of markers per cell component.",
    )

    a_pixel_b_pixel_ratio_per_cell_stats: SummaryStatistics = pydantic.Field(
        description="Summary statistics for the ratio of A-pixels to B-pixel in cell components.",
    )

    molecule_count_per_a_pixel_per_cell_stats: SummaryStatistics = pydantic.Field(
        description="Summary statistics for the number of molecules per A-pixel in cell components.",
    )

    a_pixel_count_per_b_pixel_per_cell_stats: SummaryStatistics = pydantic.Field(
        description="Summary statistics for the mean number of A-pixels per B-pixel in cell components.",
    )

    b_pixel_count_per_a_pixel_per_cell_stats: SummaryStatistics = pydantic.Field(
        description="Summary statistics for the mean number of B-pixels per A-pixel in cell components.",
    )

    # ------------------------------------------------------------------------------- #
    #   Aggregate filtering
    # ------------------------------------------------------------------------------- #

    aggregate_count: int | None = pydantic.Field(
        description="The number of components identified as aggregates."
    )

    molecules_in_aggregates_count: int | None = pydantic.Field(
        description="The total number of unique molecules in aggregates."
    )

    reads_in_aggregates_count: int | None = pydantic.Field(
        description="The total number of reads for unique molecules in aggregates."
    )

    @pydantic.computed_field(
        return_type=float,
        description=textwrap.dedent(
            """The fraction of components identified as aggregates or
            None if aggregate recovery was disabled.
            """
        ),
    )
    def fraction_aggregate_components(self) -> float | None:
        """Return the fraction of the total cell count that is marked as aggregates.

        Returns None if aggregate recovery was disabled during analysis.
        """
        if self.aggregate_count is not None and self.cell_count > 0:
            return self.aggregate_count / self.cell_count
        return None

    @pydantic.computed_field(
        return_type=float,
        description=textwrap.dedent(
            """The fraction of reads in the aggregate outliers or
            None if aggregate recovery was disabled.
            """
        ),
    )
    def fraction_reads_in_aggregates(self) -> float | None:
        """Return the fraction of reads in the aggregate outliers.

        Returns None if no aggregate recovery was disabled during analysis.
        """
        if self.reads_in_aggregates_count is not None and self.read_count > 0:
            return self.reads_in_aggregates_count / self.read_count
        return None

    @pydantic.computed_field(
        return_type=float,
        description=textwrap.dedent(
            """The fraction of molecules in the aggregate outliers or
            None if aggregate recovery was disabled.
            """
        ),
    )
    def fraction_molecules_in_aggregates(self) -> float | None:
        """Return the fraction of molecules (edges) in the aggregate outliers.

        Returns None if no aggregate recovery was disabled during analysis.
        """
        if self.molecules_in_aggregates_count is not None and self.read_count > 0:
            return self.molecules_in_aggregates_count / self.molecule_count
        return None

    # ------------------------------------------------------------------------------- #
    #   Component size filtering
    # ------------------------------------------------------------------------------- #

    min_size_threshold: Optional[int] = pydantic.Field(
        description="The minimum size threshold used for filtering components."
    )

    max_size_threshold: Optional[int] = pydantic.Field(
        description="The maximum size threshold used for filtering components."
    )

    doublet_size_threshold: Optional[int] = pydantic.Field(
        description="The doublet size threshold used for filtering components."
    )

    size_filter_fail_cell_count: int = pydantic.Field(
        description="The number of cells that do NOT pass the component size filters."
    )

    size_filter_fail_molecule_count: int = pydantic.Field(
        description="The number of molecules in components that do NOT pass the component size filters."
    )

    size_filter_fail_read_count: int = pydantic.Field(
        description="The number of reads in components that do NOT pass the component size filters."
    )



"""Functions for filtering and annotating of pixel data in edge list format.

Copyright © 2022 Pixelgen Technologies AB.
"""

import logging
from typing import Optional

import numba
import numpy as np

from pixelator.common.exceptions import PixelatorBaseException

# Work around for issue with numba and multithreading
# I got the solution from here:
# https://stackoverflow.com/questions/68131348/training-a-python-umap-model-hangs-in-a-multiprocessing-process
numba.config.THREADING_LAYER = "omp"

logger = logging.getLogger(__name__)


class NoCellsFoundException(PixelatorBaseException):
    """Raised when no cells are found in the edge list."""

    pass


def filter_components_sizes(
    component_sizes: np.ndarray,
    min_size: Optional[int],
    max_size: Optional[int],
) -> np.ndarray:
    """Filter components by size.

    Filter the component sizes provided in `component_sizes` using the size
    cut-offs defined in `min_size` and `max_size`. The components are not
    actually filtered, the function returns a boolean numpy array which
    evaluates to True if the component pass the filters.

    :param component_sizes: a numpy array with the size of each component
    :param min_size: the minimum size a component must have
    :param max_size: the maximum size a component must have
    :returns: a boolean np.array with the filtered status (True if the component
              pass the filters)
    :rtype: np.ndarray
    :raises: RuntimeError if all the components are filtered
    """
    n_components = len(component_sizes)
    logger.debug(
        "Filtering %i components using min-size=%s and max-size=%s",
        n_components,
        min_size,
        max_size,
    )

    # create a numpy array with True as default
    filter_arr = np.full((n_components), True)

    # check if any filter has been provided to the function
    if min_size is None and max_size is None:
        logger.warning("No filtering criteria provided to filter components")
    else:
        # get the components to filter (boolean array)
        if min_size is not None:
            filter_arr &= component_sizes > min_size
        if max_size is not None:
            filter_arr &= component_sizes < max_size

        # check if none of the components pass the filters
        n_components = filter_arr.sum()
        if n_components == 0:
            raise NoCellsFoundException(
                "All cells were filtered by the size filters. Consider either setting different size filters or disabling them."
            )

        logger.debug(
            "Filtering resulted in %i components that pass the filters",
            n_components,
        )
    return filter_arr



"""Collapse UMI regions independently.

Copyright © 2025 Pixelgen Technologies AB
"""

from .collapser import (
    MarkerCorrectionStats,
    RegionCollapser,
    SingleUMICollapseSampleReport,
)
from .report import IndependentCollapseSampleReport

__all__ = [
    "RegionCollapser",
    "SingleUMICollapseSampleReport",
    "MarkerCorrectionStats",
    "IndependentCollapseSampleReport",
]



"""Graph operations and graph data structures on pixelator.

Copyright © 2023 Pixelgen Technologies AB.
"""

from pixelator.common.graph.graph import Graph
from pixelator.mpx.graph.community_detection import (
    connect_components,
    recover_technical_multiplets,
)
from pixelator.mpx.graph.constants import (
    MIN_PIXELS_TO_REFINE,
)
from pixelator.mpx.graph.utils import (
    components_metrics,
    create_node_markers_counts,
    edgelist_metrics,
    update_edgelist_membership,
)

__all__ = [
    "connect_components",
    "recover_technical_multiplets",
    "MIN_PIXELS_TO_REFINE",
    "Graph",
    "components_metrics",
    "create_node_markers_counts",
    "edgelist_metrics",
    "update_edgelist_membership",
]



"""Functions used to prepare a graph for colocalization computations.

Copyright © 2023 Pixelgen Technologies AB.
"""

from typing import Literal

import pandas as pd

from pixelator.mpx.analysis.types import RegionByCountsDataFrame
from pixelator.mpx.graph.utils import Graph, create_node_markers_counts


def prepare_from_graph(graph: Graph, n_neighbours: int = 1) -> RegionByCountsDataFrame:
    """Prepare a RegionByCountsDataFrame from a graph.

    :param graph: the graph to prepare from
    :param n_neighbours: size of the neighbourhood, defaults to 1
    :return: a RegionByCountsDataFrame
    :rtype: RegionByCountsDataFrame
    """
    counts_df = create_node_markers_counts(graph=graph, k=n_neighbours)
    return counts_df


def filter_by_region_counts(
    df: RegionByCountsDataFrame, min_region_counts: int = 5
) -> RegionByCountsDataFrame:
    """Filter by counts in the region.

    Filter regions from a RegionByCountsDataFrame based on
    how many counts are in the region

    :param df: dataframe to filter
    :param min_region_counts: minimum number of counts in region > min_region_counts,
                              defaults to 5
    :return: the filtered dataframe
    :rtype: RegionByCountsDataFrame
    """
    return df[df.sum(axis="columns") > min_region_counts]


def filter_by_marker_counts(
    df: RegionByCountsDataFrame, min_marker_counts: int = 10
) -> RegionByCountsDataFrame:
    """Filter markers by minimum count.

    Filter markers from a RegionByCountsDataFrame based on how many counts
    available for that marker

    :param df: dataframe to filter
    :param min_marker_counts: marker > min_marker_counts, defaults to 10
    :return: the filtered dataframe
    :rtype: RegionByCountsDataFrame
    """
    return df.loc[:, df.sum(axis="index") > min_marker_counts]


def filter_by_unique_values(
    df: RegionByCountsDataFrame, at_least_n_unique: int = 1
) -> RegionByCountsDataFrame:
    """Filter markers by minimum number of unique values.

    Filter markers from a RegionByCountsDataFrame based on the minimum
    number of unique values for that marker

    :param df: dataframe to filter
    :param at_least_n_unique: minimum number of unique values >= at_least_n_unique,
                              defaults to 1
    :return: the filtered dataframe
    :rtype: RegionByCountsDataFrame
    """
    return df.loc[:, df.nunique() >= at_least_n_unique]



"""Module for computing colocalization statistics.

Copyright © 2023 Pixelgen Technologies AB.
"""

from typing import Tuple

import numpy as np
import pandas as pd
from sklearn.metrics.pairwise import pairwise_distances

from pixelator.mpx.analysis.colocalization.types import (
    CoLocalizationFunction,
    MarkerColocalizationResults,
)
from pixelator.mpx.analysis.types import RegionByCountsDataFrame


def _wide_correlation_matrix_to_long_correlation_results(
    df: pd.DataFrame, stat_name: str
) -> pd.DataFrame:
    # We want to keep the lower triangle, but also any value that has
    # a NaN value
    lower_triangle = np.tril(np.ones(df.shape)).astype(bool)
    df_lower_tri = df.mask(~lower_triangle)
    df_lower_tri.index.set_names(["marker_cols"], inplace=True)
    df_lower_tri.columns.set_names(["marker_rows"], inplace=True)
    correlation_values = df_lower_tri.stack(future_stack=True).dropna().reset_index()
    correlation_values.columns = ["marker_1", "marker_2", stat_name]
    correlation_values.set_index(["marker_1", "marker_2"], inplace=True)
    correlation_values.index.rename(["marker_1", "marker_2"])

    # We want to keep the NaN values in the output.
    # This adds them back again to the output (as they are dropped)
    # in the stack operation above. It's not pretty, but it's the only
    # way I've figured out to solve this.
    dx = (df.isna() & lower_triangle).values
    nn = pd.DataFrame(dx, index=df.index, columns=df.columns, dtype="object").stack()
    nn = nn[nn]
    nn[:] = pd.NA

    correlation_values_with_nan = pd.concat([correlation_values, nn], axis=0).drop(
        columns=[0]
    )
    correlation_values_with_nan.index.rename(["marker_1", "marker_2"], inplace=True)

    return correlation_values_with_nan


def _alphanumeric_sort_marker_columns(
    data: MarkerColocalizationResults,
) -> MarkerColocalizationResults:
    """Make sure that the markers are always sorted in the same order."""
    data.index = pd.MultiIndex.from_tuples(
        map(sorted, data.index.values), names=data.index.names
    )
    return data


def _drop_self_correlation(
    data: MarkerColocalizationResults,
) -> MarkerColocalizationResults:
    """Drop the self-correlation values from the data."""
    return data[data.index.get_level_values(0) != data.index.get_level_values(1)]


def pearson(df: RegionByCountsDataFrame) -> MarkerColocalizationResults:
    """Calculate the Pearson correlation between all vs all markers.

    Calculate the Pearson correlation between all vs all markers
    in the RegionByCountsDataFrame. Since these values are symmetrical only
    one of the combination of each marker pair is returned

    :param df: the RegionByCountsDataFrame to compute Pearson correlation on
    :rtype: MarkerColocalizationResults
    :return: MarkerColocalizationResults with Pearson correlations
    """
    pearson_matrix = df.corr(method="pearson")
    pearson_values = _alphanumeric_sort_marker_columns(
        _drop_self_correlation(
            _wide_correlation_matrix_to_long_correlation_results(
                pearson_matrix, "pearson"
            )
        )
    )
    return pearson_values


Pearson = CoLocalizationFunction(name="pearson", func=pearson)


def jaccard(df: RegionByCountsDataFrame) -> MarkerColocalizationResults:
    """Calculate the Jaccard index between all vs all markers.

    Calculate the Jaccard index between all vs all markers
    in the RegionByCountsDataFrame. Since these values are symmetrical only
    one of the combination of each marker pair is returned

    :param df: the RegionByCountsDataFrame to compute Jaccard indexes on
    :rtype: MarkerColocalizationResults
    :return: MarkerColocalizationResults with Jaccard indexes
    """
    jaccard_matrix = pd.DataFrame(
        1 - pairwise_distances((df.T > 0).to_numpy(dtype=bool), metric="jaccard"),
        index=df.columns.copy(),
        columns=df.columns.copy(),
    )
    jaccard_values = _alphanumeric_sort_marker_columns(
        _drop_self_correlation(
            _wide_correlation_matrix_to_long_correlation_results(
                jaccard_matrix, "jaccard"
            )
        )
    )
    return jaccard_values


Jaccard = CoLocalizationFunction(name="jaccard", func=jaccard)


def apply_multiple_stats(
    df: RegionByCountsDataFrame, funcs: Tuple[CoLocalizationFunction, ...]
) -> pd.DataFrame:
    """Compute multiple statistics on the same dataframe.

    :param df: data to compute statistics on
    :param funcs: a list of functions to use to compute the
                  statistics
    :return: a dataframe with all the statistics computed for the dataframe
    :rtype: pd.DataFrame
    """
    return pd.concat([func.func(df) for func in funcs], axis=1)



"""Model for report data returned by the single-cell amplicon stage.

Copyright © 2023 Pixelgen Technologies AB.
"""

from __future__ import annotations

import pydantic

from pixelator.mpx.report.models.base import SampleReport


class AmpliconSampleReport(SampleReport):
    """Model for data returned by the amplicon stage."""

    fraction_q30: float = pydantic.Field(
        ...,
        description="The fraction of reads with Phred score ≥ 30.",
    )

    fraction_q30_bc: float = pydantic.Field(
        ...,
        description="The fraction of reads with Phred score ≥ 30 in the barcode region.",
    )

    fraction_q30_pbs1: float = pydantic.Field(
        ...,
        description="The fraction of reads with Phred score ≥ 30 in the pixel binding site 1.",
    )

    fraction_q30_pbs2: float = pydantic.Field(
        ...,
        description="The fraction of reads with Phred score ≥ 30 in the pixel binding site 2.",
    )

    fraction_q30_umi: float = pydantic.Field(
        ...,
        description="The fraction of reads with Phred score ≥ 30 in the UMI region.",
    )

    fraction_q30_upia: float = pydantic.Field(
        ...,
        description="The fraction of reads with Phred score ≥ 30 in the UPIA region.",
    )

    fraction_q30_upib: float = pydantic.Field(
        ...,
        description="The fraction of reads with Phred score ≥ 30 in the UPIB region.",
    )



"""Functions to create a summary interactive web report.

Copyright © 2022 Pixelgen Technologies AB.
"""

import logging
from datetime import datetime
from pathlib import Path
from typing import Optional

from pixelator import __version__
from pixelator.mpx.report.common import PixelatorReporting
from pixelator.mpx.report.models.report_metadata import SampleMetadata
from pixelator.mpx.report.qcreport.builder import QCReportBuilder
from pixelator.mpx.report.qcreport.collect import collect_report_data
from pixelator.mpx.report.qcreport.types import Metrics, SampleInfo

logger = logging.getLogger("pixelator.report")


def create_dynamic_report(
    reporting: PixelatorReporting,
    sample_name: str,
    info: SampleInfo,
    output_path: Path,
) -> None:
    """Create a dynamic customer QC report for a single sample.

    A helper function to create a dynamic web report of a single sample.
    The function uses a template stored in 'report/qcreport/template.html'
    as a base template which is filled with the different metrics and stats.

    :param reporting: A PixelatorReporting object
    :param sample_name: the name of the sample
    :param info: A dictionary containing meta information about the sample
    :param output_path: the path to the output folder
    """
    sample_id = info.sample_id
    logger.info("Creating QC report for sample %s", sample_id)

    amplicon_metrics = reporting.amplicon_metrics(sample_name)
    preqc_metrics = reporting.preqc_metrics(sample_name)
    demux_metrics = reporting.demux_metrics(sample_name)
    collapse_metrics = reporting.collapse_metrics(sample_name)
    graph_metrics = reporting.graph_metrics(sample_name)
    annotate_metrics = reporting.annotate_metrics(sample_name)
    reads_flow = reporting.reads_flow(sample_name)
    molecules_flow = reporting.molecules_flow(sample_name)

    # Collect antibody metrics
    antibodies_data_values = {
        "antibody_reads": reads_flow.valid_antibody_read_count,
        "antibody_reads_usable_per_cell": annotate_metrics.read_count_per_cell_stats.mean,
        "antibody_reads_in_outliers": annotate_metrics.reads_in_aggregates_count,
        "unrecognized_antibodies": demux_metrics.unrecognised_antibody_read_count,
    }

    total_reads_per_cell = reads_flow.input_read_count / annotate_metrics.cell_count
    antibodies_data_fractions = {
        "fraction_antibody_reads": reads_flow.fraction_valid_antibody_reads,
        "fraction_antibody_reads_usable_per_cell": (
            annotate_metrics.read_count_per_cell_stats.mean / total_reads_per_cell
        ),
        "fraction_antibody_reads_in_outliers": annotate_metrics.fraction_reads_in_aggregates,
        "fraction_unrecognized_antibodies": demux_metrics.fraction_unrecognised_antibody_reads,
    }

    placeholder_cell_predictions = {
        "predicted_cell_type_b_cells": None,
        "fraction_predicted_cell_type_b_cells": None,
        "predicted_cell_type_cd4p_cells": None,
        "fraction_predicted_cell_type_cd4p_cells": None,
        "predicted_cell_type_cd8p_cells": None,
        "fraction_predicted_cell_type_cd8p_cells": None,
        "predicted_cell_type_monocytes": None,
        "fraction_predicted_cell_type_monocytes": None,
        "predicted_cell_type_nk_cells": None,
        "fraction_predicted_cell_type_nk_cells": None,
        "predicted_cell_type_unknown": None,
        "fraction_predicted_cell_type_unknown": None,
    }

    # The total number of input reads per cell.
    # Note that this is not the number of actual reads that were assigned to the cell.
    # This gives an idea of how many reads are lost in the pipeline averaged over
    # the number of cells
    average_input_read_per_cell = (
        reads_flow.input_read_count / annotate_metrics.cell_count
    )

    # Map pixelator metrics to QC report metrics
    metrics = Metrics(
        number_of_cells=annotate_metrics.cell_count,
        fraction_outlier_cells=annotate_metrics.fraction_aggregate_components,
        average_reads_usable_per_cell=int(
            annotate_metrics.read_count_per_cell_stats.mean
        ),
        average_reads_per_cell=int(average_input_read_per_cell),
        average_antibody_molecules_per_cell=annotate_metrics.molecule_count_per_cell_stats.mean,
        average_upias_per_cell=round(
            annotate_metrics.a_pixel_count_per_cell_stats.mean
        ),
        average_umis_per_upia=round(
            annotate_metrics.molecule_count_per_a_pixel_per_cell_stats.mean
        ),
        fraction_reads_in_cells=reads_flow.fraction_reads_in_cells,
        fraction_discarded_umis=molecules_flow.fraction_molecules_discarded,
        total_unique_antibodies_detected=annotate_metrics.marker_count,
        number_of_reads=reads_flow.input_read_count,
        number_of_short_reads_skipped=preqc_metrics.too_short_read_count,
        fraction_valid_pbs=reads_flow.fraction_valid_pbs_reads,
        fraction_valid_umis=reads_flow.fraction_reads_in_molecules,
        average_reads_per_molecule=graph_metrics.read_count_per_molecule_stats.mean,
        sequencing_saturation=collapse_metrics.fraction_duplicate_reads,
        fraction_q30_bases_in_antibody_barcode=amplicon_metrics.fraction_q30_bc,
        fraction_q30_bases_in_umi=amplicon_metrics.fraction_q30_umi,
        fraction_q30_bases_in_upia=amplicon_metrics.fraction_q30_upia,
        fraction_q30_bases_in_upib=amplicon_metrics.fraction_q30_upib,
        fraction_q30_bases_in_pbs1=amplicon_metrics.fraction_q30_pbs1,
        fraction_q30_bases_in_pbs2=amplicon_metrics.fraction_q30_pbs2,
        fraction_q30_bases_in_read=amplicon_metrics.fraction_q30,
        **antibodies_data_values,  # type: ignore
        **antibodies_data_fractions,  # type: ignore
        **placeholder_cell_predictions,  # type: ignore
    )

    data = collect_report_data(reporting.workdir, sample_id)

    # uses the default template
    builder = QCReportBuilder()

    report_path = str(Path(output_path) / f"{sample_id}.qc-report.html")
    with open(report_path, "wb") as f:
        builder.write(f, sample_info=info, metrics=metrics, data=data)

    logger.debug("QC report created in %s", output_path)


def make_report(
    input_path: Path,
    output_path: Path,
    panel: Optional[str],
    metadata: Optional[str],
    verbose: Optional[bool],
) -> None:
    """Parse every stage metrics and interactive plots to a single HTML file.

    This function iterates the different steps (preqc, adapterqc, demux, collapse,
    graph and annotate) for a given processed dataset (folder). The function will parse
    the metrics of all the samples present in each subfolder (step). These metrics will
    then be converted to tables. The function will also generate interactive plotly
    figures using the output data present in the graph/annotate steps. All the tables
    and figures will be saved to files and used to generate individual html report for
    each sample (web reports).

    :param input_path: the path to results folder containing all the steps
    :param output_path: the path to the output folder
    :param panel: a path to a panel file or a panel name to load from the config
    :param metadata: a path to a metadata file
    :param verbose: run if verbose mode when true
    :returns: None
    :rtype: None
    :raises AssertionError: when the input is not valid
    :raises RuntimeError: when the metadata is not valid
    """
    from pixelator.mpx.config import config, load_antibody_panel

    logger.info("Generating QC reports from %s", input_path)

    # collect and save metrics
    if panel is not None:
        panel_obj = load_antibody_panel(config, panel)
    else:
        panel_obj = None

    reporting = PixelatorReporting(Path(input_path))

    sample_metadata = SampleMetadata.from_csv(metadata) if metadata else None

    # create the dynamic report (one per sample)

    for sample in reporting.samples():
        this_sample_metadata = (
            sample_metadata.get_by_id(sample) if sample_metadata else None
        )
        sample_description = (
            this_sample_metadata.description if this_sample_metadata else ""
        )

        invocation_info = reporting.cli_invocation_info(sample)
        pixel_version = ""

        try:
            collapse_design_option = invocation_info.get_option("collapse", "--design")
            pixel_version = collapse_design_option.value
        except KeyError:
            logger.warning("No design info found for sample %s", sample)

        webreport_panel_name = None
        webreport_panel_version = None

        if panel_obj is not None:
            webreport_panel_name = (
                panel_obj.name if panel_obj.name else str(panel_obj.filename)
            )
            webreport_panel_version = panel_obj.version

        # create SampleInfo object
        sample_info = SampleInfo(
            pixelator_version=__version__,
            generation_date=datetime.utcnow().isoformat(),
            sample_id=sample,
            sample_description=sample_description,
            pixel_version=pixel_version,
            panel_name=webreport_panel_name,
            panel_version=webreport_panel_version,
            parameters=list(invocation_info),
        )

        create_dynamic_report(
            reporting,
            sample_name=sample,
            info=sample_info,
            output_path=output_path,
        )

    logger.info("Finished generating QC reports")



"""Denoise module.

Functions and classes relating to denoising PNA components.

Copyright © 2025 Pixelgen Technologies AB
"""

import logging
import random
import tempfile
from itertools import chain
from pathlib import Path

import duckdb
import networkx as nx
import numpy as np
import pandas as pd
import polars as pl
from scipy.stats import fisher_exact

from pixelator.common.annotate.aggregates import call_aggregates
from pixelator.pna.analysis_engine import PerComponentTask
from pixelator.pna.anndata import pna_edgelist_to_anndata
from pixelator.pna.config import pna_config
from pixelator.pna.config.panel import PNAAntibodyPanel, load_antibody_panel
from pixelator.pna.graph import PNAGraph
from pixelator.pna.pixeldataset import PNAPixelDataset, PxlFile
from pixelator.pna.pixeldataset.io import PixelFileWriter

logger = logging.getLogger(__name__)


def _calculate_core_marker_counts(
    node_marker_counts: pd.DataFrame, node_core_numbers: pd.Series
) -> pd.DataFrame:
    """Calculate marker counts for one-core and higher-core nodes."""
    one_core_counts = node_marker_counts[
        (node_core_numbers[node_marker_counts.index] == 1).values
    ].sum()
    higher_core_counts = node_marker_counts[
        (node_core_numbers[node_marker_counts.index] > 1).values
    ].sum()

    marker_counts = (
        one_core_counts.to_frame(name="one_core")
        .join(higher_core_counts.to_frame(name="higher_core"), how="outer")
        .fillna(0)
    )
    return marker_counts


def _perform_fishers_exact_test(
    marker_counts: pd.DataFrame,
    pval_significance_threshold: float,
) -> list:
    """Perform Fisher's exact test to identify over-expressed markers."""
    overexpressed_markers = []
    total_one_core_markers = marker_counts["one_core"].sum()
    total_higher_core_markers = marker_counts["higher_core"].sum()
    for marker in marker_counts.index:
        contingency_mat = [
            [
                marker_counts.loc[marker, "one_core"],
                total_one_core_markers - marker_counts.loc[marker, "one_core"],
            ],
            [
                marker_counts.loc[marker, "higher_core"],
                total_higher_core_markers - marker_counts.loc[marker, "higher_core"],
            ],
        ]
        pval = fisher_exact(contingency_mat, alternative="greater").pvalue
        if pval < pval_significance_threshold:
            overexpressed_markers.append(marker)
    return overexpressed_markers


def _calculate_excess_counts(
    marker_counts: pd.DataFrame,
    overexpressed_markers: list,
    inflate_factor: float,
) -> list:
    """Calculate inflated excess counts for over-expressed markers."""
    results = []
    total_one_core_markers = marker_counts["one_core"].sum()
    total_higher_core_markers = marker_counts["higher_core"].sum()
    for marker in overexpressed_markers:
        expected_count = np.round(
            total_one_core_markers
            * marker_counts.loc[marker, "higher_core"]
            / total_higher_core_markers
        )
        excess = int(
            np.ceil(
                inflate_factor
                * (marker_counts.loc[marker, "one_core"] - expected_count)
            )
        )
        results.append((marker, excess))
    return results


def get_overexpressed_markers_in_one_core(
    node_marker_counts: pd.DataFrame,
    node_core_numbers: pd.Series,
    pval_significance_threshold: float = 0.05,
    inflate_factor: float = 1.5,
) -> pd.DataFrame:
    """Identify over-expressed markers in nodes belonging to core 1.

    This function calculates the over-expression of markers in nodes that belong
    to core 1 by comparing their marker counts to nodes in higher cores. It
    uses Fisher's exact test to determine statistical significance and inflates
    the excess count of markers based on the provided inflate factor.
    The inflate factor increases the number of nodes with a given over-expressed
    marker to be removed from the one-core layer.

    Args:
        node_marker_counts (pd.DataFrame): A DataFrame where rows represent
            nodes and columns represent markers, with values indicating the
            count of each marker in each node.
        node_core_numbers (pd.Series): A Series where the index corresponds to
            the nodes and the values indicate the core number each node belongs to.
        pval_significance_threshold (float, optional): The p-value threshold
            for statistical significance in Fisher's exact test. Defaults to 0.05.
        inflate_factor (float, optional): A factor used to inflate the excess
            count of markers identified as overexpressed. Defaults to 1.5.
        one_core_ratio_threshold(float, optional): Components with higher nodes in
                                                   their one-core layer are not denoised.

    Returns:
        pd.DataFrame: A DataFrame with two columns:
            - "name": The names of the overexpressed markers.
            - "count": The inflated excess count of each overexpressed marker.

    """
    marker_counts = _calculate_core_marker_counts(node_marker_counts, node_core_numbers)

    overexpressed_markers = _perform_fishers_exact_test(
        marker_counts,
        pval_significance_threshold,
    )

    results = _calculate_excess_counts(
        marker_counts,
        overexpressed_markers,
        inflate_factor,
    )

    return pd.DataFrame(results, columns=["name", "count"])


def _sample_nodes_to_be_removed(
    node_marker_counts: pd.DataFrame, markers_to_remove: pd.DataFrame
) -> list:
    to_be_removed = []
    for _, marker in markers_to_remove.iterrows():
        marker_nodes = list(
            node_marker_counts[node_marker_counts[marker["name"]] > 0].index
        )
        marker_count_available = len(marker_nodes)
        n_to_be_removed = min(marker_count_available, marker["count"])
        rand_gen = random.Random(0)
        to_be_removed += list(rand_gen.sample(marker_nodes, n_to_be_removed))
    return to_be_removed


def get_stranded_nodes(component: PNAGraph, nodes_to_remove: list = []) -> list:
    """Identify nodes that become stranded after removing nodes_to_remove.

    Args:
        component (PNAGraph): The graph component from which nodes will be removed.
        nodes_to_remove (list): A list of nodes to be removed from the graph.

    Returns:
        list: A list of stranded nodes that are disconnected from the largest
        connected component after nodes_to_remove are removed.

    """
    graph = component.raw.copy()
    graph.remove_nodes_from(nodes_to_remove)
    connected_components = sorted(nx.connected_components(graph), key=len, reverse=True)
    stranded_nodes = list(chain.from_iterable(connected_components[1:]))
    return stranded_nodes


def denoise_one_core_layer(
    component: PNAGraph,
    pval_significance_threshold: float = 0.05,
    inflate_factor: float = 1.5,
    one_core_ratio_threshold: float = 0.9,
) -> list:
    """Identify and remove markers over-expressed in the one-core layer of a graph.

    This function analyzes the one-core layer of a graph, identifies markers
    that are over-expressed using a statistical significance threshold, and
    removes nodes associated with those markers. Additionally, it ensures
    that stranded nodes (nodes disconnected from the graph due to removal)
    are also removed.

    Args:
        component (PNAGraph): The graph component to process, containing
            node marker counts and raw graph data.
        pval_significance_threshold (float, optional): The p-value threshold
            for determining marker overexpression significance. Defaults to 0.05.
        inflate_factor (float, optional): A factor used for inflating certain
            calculations (not explicitly used in the provided code). Defaults to 1.5.
        one_core_ratio_threshold(float, optional): Components with higher nodes in
                                                   their one-core layer are not denoised.

    Returns:
        list: A list of nodes to be removed from the one-core layer of the graph.

    """
    node_marker_counts = component.node_marker_counts
    node_core_numbers = pd.Series(nx.core_number(component.raw))
    if (node_core_numbers <= 1).mean() >= one_core_ratio_threshold:
        logger.debug(
            "Too many low core number nodes. Skipping denoising for this component."
        )
        return [None]  # Marking component as unqualified for denoising
    markers_to_remove = get_overexpressed_markers_in_one_core(
        node_marker_counts=node_marker_counts,
        node_core_numbers=node_core_numbers,
        pval_significance_threshold=pval_significance_threshold,
    )
    one_core_layer = node_marker_counts[
        ((node_core_numbers[node_marker_counts.index] == 1).values)
    ]
    nodes_to_remove = _sample_nodes_to_be_removed(one_core_layer, markers_to_remove)
    stranded_nodes = get_stranded_nodes(component, nodes_to_remove)
    nodes_to_remove += stranded_nodes
    return nodes_to_remove


def write_denoised_edgelist(
    pxl: PNAPixelDataset, umis_to_remove: list, output_edgelist_path: str
):
    """Write a filtered edgelist (with specified UMIs removed) to a Parquet file.

    This function takes an existing PNAPixelDataset, removes edges associated
    with the specified UMIs, and writes the resulting edgelist to a Parquet file
    at the specified path. It does not create a new pixel (.pxl) file or update
    any AnnData object.

    Args:
        pxl (PNAPixelDataset): The original pixel dataset containing the edgelist.
        umis_to_remove (list): A list of UMIs (nodes) to be removed from the edgelist.
        output_edgelist_path (str): The file path where the filtered edgelist Parquet file will be saved.

    """
    with pxl.view as con:
        con.execute(
            f"""
            COPY(
                SELECT *
                FROM edgelist
                WHERE umi1 NOT IN (SELECT UNNEST(?))
                AND umi2 NOT IN (SELECT UNNEST(?))
            ) TO '{output_edgelist_path}' (FORMAT PARQUET)
            """,
            [umis_to_remove, umis_to_remove],
        )


class DenoiseOneCore(PerComponentTask):
    """Denoise bleed-over markers in parts of the component with low coreness."""

    TASK_NAME = "denoise-one-core"

    def __init__(
        self,
        pval_significance_threshold: float = 0.05,
        inflate_factor: float = 1.5,
        one_core_ratio_threshold: float = 0.9,
    ):
        """Initialize a DenoiseOneCore instance.

        Args:
            pval_significance_threshold (float): The p-value threshold for considering
            a marker over-expressed in the one-core layer.
            inflate_factor: A factor used to inflate the excess count of markers.
            one_core_ratio_threshold: Components with higher nodes in their one-core
                                      layer are not denoised.

        """
        self.pval_significance_threshold = pval_significance_threshold
        self.inflate_factor = inflate_factor
        self.one_core_ratio_threshold = one_core_ratio_threshold

    def run_on_component_graph(
        self, component: PNAGraph, component_id: str
    ) -> pd.DataFrame:
        """Execute one-core denoising on a given component graph and return nodes to remove.

        This function performs denoising on the provided PNAGraph component by
        identifying nodes to be removed based on a single core layer denoising
        process. The resulting nodes are returned in a DataFrame along with
        their associated component ID.

        Args:
            component (PNAGraph): The graph component to be denoised.
            component_id (str): The identifier for the graph component.

        Returns:
            pd.DataFrame: A DataFrame containing the nodes to be removed with the following columns:
                - "umi": The unique identifier of the node to be removed.
                - "component": The ID of the component the node belongs to.

        """
        logger.debug(f"Running low-core denoising on component {component_id}")
        nodes_to_remove = pd.DataFrame(
            denoise_one_core_layer(
                component,
                pval_significance_threshold=self.pval_significance_threshold,
                inflate_factor=self.inflate_factor,
                one_core_ratio_threshold=self.one_core_ratio_threshold,
            ),
            columns=["umi"],
        )
        nodes_to_remove["component"] = component_id
        return nodes_to_remove

    def add_to_pixel_file(self, data: pd.DataFrame, pxl_file_target: PxlFile):
        """Add denoised component to a pixel file by updating the edgelist and adata.

        This function reads an existing pixel file, filters its edgelist
        based on the provided data, and updates the file with the modified
        edgelist and corresponding AnnData object.

        Args:
            data (pd.DataFrame): A DataFrame containing the data to be used
                for filtering.
            pxl_file_target (PxlFile): The target pixel file to which
                the data will be added.

        """
        pxl = PNAPixelDataset.from_files(pxl_file_target)
        try:
            panel = PNAAntibodyPanel.from_pxl(pxl_file_target.path)
        except KeyError:
            # If pxl file does not contain panel data, try to load it from
            # the panel name.
            # This will happen when old pxl files generated before v0.22.0
            # are used.
            panel_name = pxl.metadata().popitem()[1]["panel_name"]
            panel = load_antibody_panel(pna_config, panel_name)
        nodes_to_remove = (
            data.loc[~data["umi"].isna(), "umi"].astype(np.uint64).tolist()
        )

        with tempfile.TemporaryDirectory() as temp_dir:
            denoised_edgelist_path = temp_dir + "/denoised_edgelist.parquet"
            write_denoised_edgelist(pxl, nodes_to_remove, denoised_edgelist_path)
            with PixelFileWriter(pxl_file_target.path) as writer:
                writer.write_edgelist(Path(denoised_edgelist_path))
                adata = pna_edgelist_to_anndata(writer.get_connection(), panel)
                call_aggregates(adata)
                denoise_info = pd.DataFrame(index=adata.obs.index)
                denoise_info["disqualified_for_denoising"] = False
                denoise_info.loc[
                    data.loc[data["umi"].isna(), "component"],
                    "disqualified_for_denoising",
                ] = True
                n_umis_removed = (
                    data.loc[~data["umi"].isna(), :].groupby("component").size()
                )
                denoise_info["number_of_nodes_removed_in_denoise"] = n_umis_removed
                denoise_info["number_of_nodes_removed_in_denoise"] = denoise_info[
                    "number_of_nodes_removed_in_denoise"
                ].fillna(0)
                adata.obs = adata.obs.join(denoise_info, how="left")

                writer.write_adata(adata)
